/* AC3 Mesquite */
/* FIC report ~~ Copyright 2002-2006 P. Midford*/
package mesquite.ac3.FICReport;
/*~~  */

import mesquite.lib.*;
import mesquite.lib.characters.*;
import mesquite.lib.duties.CharSourceCoordObed;
import mesquite.parsimony.lib.ParsAncStatesForModel;
import mesquite.categ.lib.CategoricalDistribution;
import mesquite.cont.lib.*;
import mesquite.ac3.lib.*;

/* ============  a file dumper for .FIC files created by PDTREE (Garland et al. 1993) ============*/

//FIC files contain the following columns:

//1.  Name of Contrast as Node - Node (e.g., 'Tm - Ta');
//2.  Name of Node at Base of Contrast
//3.  Unstandardized Contrast for Trait 1
//4.  Unstandardized Contrast for Trait 2
//5.  Standard Deviation of Contrast
//= square root of sum of corrected branch lengths
//6.  Nodal Value for Trait 1
//7.  Nodal Value for Trait 2
//8.  Uncorrected Height of Base Node from Root of Tree
//9.  Corrected Height of Base Node from Root of Tree
//(as used in screens 5, 6, 7, 8, and 11)
//10.  Average Uncorrected Height of Daughter Nodes
//11.  Average Corrected Height of Daughter Nodes
//12.  Name of First Daughter Branch (Descendant-Ancestor)
//13.  Corrected Length of First Daughter Branch
//14.  Name of Second Daughter Branch (Descendant-Ancestor)
//15.  Corrected Length of Second Daughter Branch


/** */
public class FICReport extends AC3FileReporter {
	ContrastForChar contrastsTask;
	StatPakSource statPakTask;
	private NumberArray xArray;
	private NumberArray yArray;
	private NumberArray xRawContrastArray;
	private NumberArray yRawContrastArray;
	private NumberArray SDArray;
	private NumberArray VPrimeArray;
	private NumberArray heightArray;
	private NumberArray rawHeightArray;
	private final String processedSuffixes = "#-+*:.=!";
	private int nameSize;
	private NumberArray nodesToRecalc;
	private String [] statesOfNodes;
	private double [] contrastX;
	private double [][] contrastY;


	/*.................................................................................................................*/
	public boolean startJob(String arguments, Object condition, CommandRecord commandRec, boolean hiredByName) {
		if (contrastsTask == null)
			contrastsTask =  (ContrastForChar)hireNamedEmployee(commandRec, ContrastForChar.class, "#FelsContrWithChar");
		if (contrastsTask == null) {
			alert(getName() + " couldn't set up the FIC report generator because no calculator of Felsenstein's contrasts obtained.");
			return false;
		}
		
		xArray = new NumberArray(3);
		yArray = new NumberArray(3);
		xRawContrastArray = new NumberArray(3);
		yRawContrastArray = new NumberArray(3);
		SDArray = new NumberArray(3);
		VPrimeArray = new NumberArray(3);
		heightArray = new NumberArray(3);
		rawHeightArray = new NumberArray(3);
		nodesToRecalc = new NumberArray(3);
		return true;  //false if no calculating employees could be found
	}

	public void employeeQuit(MesquiteModule m){
		if (m==contrastsTask)
			iQuit();
	}

	/*.................................................................................................................*/
	/** Always true, because an FIC file always has tables. */
	public boolean hasTables() {
		return true;
	}

	/*.................................................................................................................*/
	/** ask for formatting options */
	public boolean getReportOptions(boolean dataSelected, boolean taxaSelected){
		MesquiteInteger buttonPressed = new MesquiteInteger(1);
		AC3ReporterDialog reportDialog = new AC3ReporterDialog(this,containerOfModule(), "FIC Report Options", buttonPressed);
		reportDialog.addDefaultPanels(dataSelected, taxaSelected,hasTables());
		boolean ok = (reportDialog.query(dataSelected,taxaSelected,convertSpaces)==0);
		reportDialog.dispose();
		return ok;
	}

	/*.................................................................................................................*/
	// This procedure tries to get a name from somewhere, and if it appears to be processed, unprocesses it
	private String getName(Tree tree, int node) {
		String name;
		if (tree.getNodeLabel(node) == null)  { // No name, so we'll generate one from the node number
			name = "Node" + Integer.toString(node);
		}
		else {
			name = tree.getNodeLabel(node);
			if (((name.length() == 2) && (processedSuffixes.indexOf(name.charAt(1)) != -1)) ||   //guessing name is processed
					((name.length() == 3) && (processedSuffixes.indexOf(name.charAt(2)) != -1))) {
				name = name.substring(0,name.length()-1);
			}
		}
		return name;
	}
	/*.................................................................................................................*/
	// This finds the longest node label on the tree
	private int checkNameLengths(Tree tree,int node) {
		String name = getName(tree,node);
		int size = name.length();
		if (tree.nodeIsInternal(node)) { 
			for (int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter)) {
				int daughterLength = checkNameLengths(tree,daughter);
				if (size < daughterLength)
					size = daughterLength;
			}
		}
		return size;
	}
	/*.................................................................................................................*/
	// This generates a fill string consisting of spaces equal to the difference between the length of name and the value len
	private String addFill(String name, int len) {
		if (name.length() == len)  //is this name construct equal to the maximum length?
			return "";  // no fill needed
		else {
			StringBuffer buf = new StringBuffer(len-name.length());
			for (int i=name.length();i<len;i++)
				buf.append(' ');
			return buf.toString();
		}
	}

	/*...............................................................*/
	//This should export the 15 columns in the .FIC file.  Like PDI, this assumes all nodes are fully resolved and
	//have exactly two daugthers.  Unlike PDI, there is no obvious best way to extend to more than two characters -
	//the raw contrasts are currently listed in columns 3&4, so would additional characters go on the end (16,17,etc)
	//inserted after 5 and 6.  Since these files are intended to go into SPSS and similar programs, extra columns 
	//anywhere might break things.
	private void reportNode(Tree tree, StringBuffer outputBuffer, int node, String lineEnding, String tableDelimiter) {
		if (tree.nodeIsInternal(node)) { // only internal nodes have contrasts
			String myName;
			String firstDName;
			String secondDName;
			int FirstDaughter;
			int SecondDaughter;
			int position=0;
			boolean recalculatedNode =false;
			if(nodesToRecalc!=null)
			for (int i=0;i<nodesToRecalc.getSize();i+=3){
				if(nodesToRecalc.getInt(i)!=0){
					if (nodesToRecalc.getInt(i)==node){
						recalculatedNode=true;
						position=i;
					}
				}
				else
					break;
			}
			if (recalculatedNode){
				FirstDaughter = (int)nodesToRecalc.getDouble(position+1);
				SecondDaughter = (int)nodesToRecalc.getDouble(position+2);
			}
			else{
				FirstDaughter = tree.firstDaughterOfNode(node);
				SecondDaughter = tree.lastDaughterOfNode(node);
			}

			if (convertSpaces){
				myName = spaceConvertString(getName(tree,node));
				firstDName = spaceConvertString(getName(tree,FirstDaughter));
				secondDName = spaceConvertString(getName(tree,SecondDaughter));
			}
			else {
				myName = getName(tree,node);
				firstDName = getName(tree,FirstDaughter);
				secondDName = getName(tree,SecondDaughter);
			}
			double Daughter1State1 = xArray.getDouble(FirstDaughter);
			double Daughter1State2 = yArray.getDouble(FirstDaughter);
			double Daughter2State1 = xArray.getDouble(SecondDaughter);
			double Daughter2State2 = yArray.getDouble(SecondDaughter);

			//     1.  Name of Contrast as Node - Node
			String tmp = firstDName + "-" + secondDName; 
			outputBuffer.append(tmp + addFill(tmp,2*nameSize+1) +  tableDelimiter);
			//     2.  Name of Node at Base of Contrast
			outputBuffer.append(myName + addFill(myName,nameSize) + tableDelimiter);
			//     3.  Unstandardized Contrast for Trait 1
			if (MesquiteDouble.isCombinable(Daughter1State1) &&
					MesquiteDouble.isCombinable(Daughter2State1)) {
				outputBuffer.append(MesquiteDouble.toFixedWidthString((Daughter1State1-Daughter2State1),15,false) + tableDelimiter);
			}
			else outputBuffer.append("-99999          " + tableDelimiter);  //AC3 style undefined value
			//     4.  Unstandardized Contrast for Trait 2
			if (MesquiteDouble.isCombinable(Daughter1State1) &&
					MesquiteDouble.isCombinable(Daughter2State1)) {
				outputBuffer.append(MesquiteDouble.toFixedWidthString((Daughter1State2-Daughter2State2),15,false) + tableDelimiter);
			}
			else outputBuffer.append("-99999          " + tableDelimiter);  //AC3 style undefined value
			//     5.  Standard Deviation of Contrast
			//             = square root of sum of corrected branch lengths
			if (MesquiteDouble.isCombinable(SDArray.getDouble(node)))
				outputBuffer.append(MesquiteDouble.toFixedWidthString(SDArray.getDouble(node),15,false) + tableDelimiter);
			else outputBuffer.append("-99999          " + tableDelimiter);  //AC3 style undefined value
			//     6.  Nodal Value for Trait 1
		//	if (MesquiteDouble.isCombinable(xArray.getDouble(node)))
				if (false)
				outputBuffer.append(MesquiteDouble.toFixedWidthString(xArray.getDouble(node),15,false) + tableDelimiter);
		//	else outputBuffer.append("-99999          " + tableDelimiter);  //AC3 style undefined value
			else outputBuffer.append(statesOfNodes[node]+ "          " + tableDelimiter);  //AC3 style undefined value
			//     7.  Nodal Value for Trait 2
			if (MesquiteDouble.isCombinable(yArray.getDouble(node)))
				outputBuffer.append(MesquiteDouble.toFixedWidthString(yArray.getDouble(node),15,false) + tableDelimiter);
			else outputBuffer.append("-99999          " + tableDelimiter);  //AC3 style undefined value
			//     8.  Uncorrected Height of Base Node from Root of Tree
			if (MesquiteDouble.isCombinable(rawHeightArray.getDouble(node)))
				outputBuffer.append(MesquiteDouble.toFixedWidthString(rawHeightArray.getDouble(node),15,false) + tableDelimiter);
			else outputBuffer.append("-99999          " + tableDelimiter);  //AC3 style undefined value
			//     9.  Corrected Height of Base Node from Root of Tree
			//		(as used in screens 5, 6, 7, 8, and 11)
			if (MesquiteDouble.isCombinable(heightArray.getDouble(node)))
				outputBuffer.append(MesquiteDouble.toFixedWidthString(heightArray.getDouble(node),15,false) + tableDelimiter);
			else outputBuffer.append("-99999          " + tableDelimiter);  //AC3 style undefined value
			//    10.  Average Uncorrected Height of Daughter Nodes
			if (MesquiteDouble.isCombinable(rawHeightArray.getDouble(FirstDaughter)) &&
					MesquiteDouble.isCombinable(rawHeightArray.getDouble(SecondDaughter))) {
				outputBuffer.append(MesquiteDouble.toFixedWidthString((rawHeightArray.getDouble(FirstDaughter) + rawHeightArray.getDouble(SecondDaughter) )/2,15,false) + tableDelimiter);
			}
			else outputBuffer.append("-99999          " + tableDelimiter);  //AC3 style undefined value
			//    11.  Average Corrected Height of Daughter Nodes
			if (MesquiteDouble.isCombinable(heightArray.getDouble(FirstDaughter)) &&
					MesquiteDouble.isCombinable(heightArray.getDouble(SecondDaughter))) {
				outputBuffer.append(MesquiteDouble.toFixedWidthString((heightArray.getDouble(FirstDaughter) + heightArray.getDouble(SecondDaughter) )/2,15,false) + tableDelimiter);
			}
			else outputBuffer.append("-99999          " + tableDelimiter);  //AC3 style undefined value
			//    12.  Name of First Daughter Branch (Descendant-Ancestor)
			tmp = firstDName + "-" + myName; 
			outputBuffer.append(tmp + addFill(tmp,2*nameSize+1) + tableDelimiter);
			//    13.  Corrected Length of First Daughter Branch
			if (MesquiteDouble.isCombinable(VPrimeArray.getDouble(FirstDaughter)))
				outputBuffer.append(MesquiteDouble.toFixedWidthString(VPrimeArray.getDouble(FirstDaughter),15,false) + tableDelimiter);
			else outputBuffer.append("-99999          " + tableDelimiter);  //AC3 style undefined value
			//    14.  Name of Second Daughter Branch (Descendant-Ancestor)
			tmp = secondDName + "-" + myName; 
			outputBuffer.append(tmp + addFill(tmp,2*nameSize+1) + tableDelimiter);
			//    15.  Corrected Length of Second Daughter Branch
			if (MesquiteDouble.isCombinable(VPrimeArray.getDouble(SecondDaughter)))
				outputBuffer.append(MesquiteDouble.toFixedWidthString(VPrimeArray.getDouble(SecondDaughter),15,false));
			else outputBuffer.append("-99999          " + tableDelimiter);  //PDAP style undefined value

			outputBuffer.append(lineEnding);
		}
		for (int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter))
			reportNode(tree, outputBuffer, daughter,lineEnding,tableDelimiter);
	}

	/*.................................................................................................................*/
	// This writes the string to a file.
	public void writeReport(MesquiteFile file,CharacterDistribution data1,CharacterDistribution data2,Tree theTree, String arguments, CommandRecord commandRec) { //todo: this should be passed scripting

		String output = reportToString(data1,data2,theTree,commandRec);
		if (!(output.equalsIgnoreCase(""))) {
			String name = getProject().getHomeFileName();
			if (name==null)
				name = "untitled.fic";
			else 
				name = stripNex(name) + ".fic";
			saveReportFile(output,arguments,name,commandRec);

		}
	}

	/**
	 * @return String contains the report
	 * No line or column delimiter specified, puts up a window instead.
	 */
	public String reportToString(CharacterDistribution data1,CharacterDistribution data2,Tree theTree, CommandRecord commandRec) { //todo: this should be passed scripting
		if (!getReportOptions(false,false))
			return "";
		return reportToString(data1,data2,theTree,getLineEnding(),getTableDelimiter(),commandRec);
	}
	
	public String reportContrastToString(CharacterDistribution char1,CharSourceCoordObed char2,Tree theTree, CommandRecord commandRec) { //todo: this should be passed scripting
		if (!getReportOptions(false,false))
			return "";
		return reportContrastToString(char1,char2,theTree,getLineEnding(),getTableDelimiter(),commandRec);
	}

	public String reportContrastToString(CharacterDistribution data1,CharSourceCoordObed data2,Tree theTree,String lineDelimiter, String columnDelimiter, CommandRecord commandRec) { //todo: this should be passed scripting
		long max=data2.getNumberOfCharacters(theTree.getTaxa(), CommandRecord.nonscriptingRecord);
		int numNodes = theTree.getNumNodeSpaces();
		NumberArray yContrasts=yRawContrastArray.cloneArray();
		yContrasts.resetSize(numNodes);
		StringBuffer outputBuffer = new StringBuffer(256*(data1.getTaxa().getNumTaxa()+3) + 500);
		CategoricalDistribution states1=(CategoricalDistribution)data1;
		AC3ContrastCalculator categoricalCalculator=new AC3ContrastCalculator(theTree,states1);
		double Xcontrasts[]=categoricalCalculator.contrastCalculation();
		nodesToRecalc=categoricalCalculator.getNodesToRecalculate();
		contrastX=new double [Xcontrasts.length];
		contrastY=new double [(int)max][Xcontrasts.length];
		boolean std= columnDelimiter=="\t";
		if(!std)
		for (int j=0;j<Xcontrasts.length;j++)
			contrastX[j]=Xcontrasts[j];
		else{
			SDArray.resetSize(numNodes);
		if(nodesToRecalc!=null)
			SDArray=nodesToRecalc.cloneArray();
		contrastsTask.setOption(ContrastForChar.SD,CommandRecord.nonscriptingRecord);
		contrastsTask.calculateNumbers(theTree, (ContinuousDistribution)data2.getCharacter(theTree, 0, CommandRecord.nonscriptingRecord), SDArray, null, CommandRecord.nonscriptingRecord);
		for (int j=0;j<Xcontrasts.length;j++)
			if(!MesquiteDouble.isUnassigned(Xcontrasts[j]) && SDArray.getDouble(j)!=0)
				contrastX[j]=Xcontrasts[j]/SDArray.getDouble(j);
			else 
				contrastX[j]=Xcontrasts[j];
		}
		contrastsTask.setOption(ContrastForChar.CONTRAST, CommandRecord.nonscriptingRecord);
		for (int i=0;i<max;i++){
			ContinuousDistribution states2=(ContinuousDistribution)data2.getCharacter(theTree, i, CommandRecord.nonscriptingRecord);
		if(nodesToRecalc!=null)
			yContrasts=nodesToRecalc.cloneArray();
		contrastsTask.calculateNumbers(theTree, states2, yContrasts, null, CommandRecord.nonscriptingRecord);
		for (int k=0;k<yContrasts.getSize();k++)
			contrastY[i][k]=yContrasts.getDouble(k);
						}
		
		reportContrast(theTree,outputBuffer,theTree.getRoot(),lineDelimiter,"\t",std);

		outputBuffer.append("\n");
		return outputBuffer.toString();	
	}
	
private void reportContrast(Tree tree, StringBuffer outputBuffer, int node, String lineEnding, String tableDelimiter,boolean standardized) {
	if(!standardized)
	outputBuffer.append("Contrasts of X and Y\n");
	else 
		outputBuffer.append("Contrasts of X(standardized) and Y\n");
	outputBuffer.append("X"+tableDelimiter);
	for (int i =0;i<contrastY.length;i++){
	outputBuffer.append("Y"+i+tableDelimiter);
	}
	outputBuffer.append("\n");
	for (int j=0;j<contrastX.length;j++){
		if(!MesquiteDouble.isUnassigned(contrastX[j]) && contrastX[j]>=0){
		outputBuffer.append(contrastX[j] +tableDelimiter);
		for (int i =0;i<contrastY.length;i++){
			outputBuffer.append(contrastY[i][j]+tableDelimiter);
			}
		outputBuffer.append(lineEnding);
		}
		if(!MesquiteDouble.isUnassigned(contrastX[j]) && contrastX[j]<0){
			outputBuffer.append(-1*contrastX[j] +tableDelimiter);
			for (int i =0;i<contrastY.length;i++){
				outputBuffer.append(-1*contrastY[i][j]+tableDelimiter);
				}
			outputBuffer.append(lineEnding);
			}
	}
}
	
	/*.................................................................................................................*/
	// This actually writes the report.
	public String reportToString(CharacterDistribution data1,CharacterDistribution data2,Tree theTree,String lineDelimiter, String columnDelimiter, CommandRecord commandRec) { //todo: this should be passed scripting
		if ((!(data1 instanceof CategoricalDistribution))  || !(data2 instanceof ContinuousDistribution)) {
			fireEmployee(contrastsTask);
			contrastsTask = null;		
			return "";
		}

		CategoricalDistribution states1 = (CategoricalDistribution)data1;
		ContinuousDistribution states2 = (ContinuousDistribution)data2;
		if (data1.getTaxa().getNumTaxa() != data2.getTaxa().getNumTaxa()) { // differing taxa sets (can we just compare the Taxa objects?) very bad
			fireEmployee(contrastsTask);
			contrastsTask = null;
			return "";
		}
		Taxa taxa = data1.getTaxa();
		int numTaxa = taxa.getNumTaxa();
		AC3ContrastCalculator categoricalCalculator=new AC3ContrastCalculator(theTree,states1);
		// Here we check whether this tree follows the rules for pdi files: internal nodes have names, all names are <= 2 chars or processed, ??bifurcating with ZLB's for polytomies...
		nameSize = checkNameLengths(theTree,theTree.getRoot());
		//               int numChars = data.getNumChars();
//		TreeSource source = 
//		(TreeSource)hireNamedEmployee(CommandRecord.nonscriptingRecord,TreeSource.class, "#StoredTrees");
//		int whichTree = source.queryUserChoose(taxa, "to include in FIC file for PDAP", CommandRecord.nonscriptingRecord);
//		Tree tree = source.getTree(taxa,whichTree, CommandRecord.nonscriptingRecord);
//		fireEmployee(source);
		StringBuffer outputBuffer = new StringBuffer(256*(numTaxa+3) + 500);
//		int char1 =0;  // defaults
//		int char2 =1;
//		CharacterDistribution charStatesX = data.getCharacterDistribution(char1);
//		CharacterDistribution charStatesY = data.getCharacterDistribution(char2);

		//outputBuffer.append(numTaxa + getLineEnding() + getLineEnding());

		int numNodes = theTree.getNumNodeSpaces();
		xRawContrastArray.resetSize(numNodes);
		if (contrastsTask == null) {
			alert(getName() + " FIC report generator exiting because the calculator of Felsenstein's contrasts is gone.");
			iQuit();
		}
		contrastsTask.setOption(ContrastForChar.RAWCONTRAST, CommandRecord.nonscriptingRecord);

		//contrastsTask.calculateNumbers(theTree, states1, xRawContrastArray, null, CommandRecord.nonscriptingRecord);
		double contrasts[]=categoricalCalculator.contrastCalculation();
		for (int i=0;i<contrasts.length;i++)
			xRawContrastArray.setValue(i, contrasts[i]);
		nodesToRecalc=categoricalCalculator.getNodesToRecalculate();
		yRawContrastArray.resetSize(numNodes);
		contrastsTask.calculateNumbers(theTree, states2, yRawContrastArray, null, CommandRecord.nonscriptingRecord);
		SDArray.resetSize(numNodes);
		contrastsTask.setOption(ContrastForChar.SD,CommandRecord.nonscriptingRecord);
		if (nodesToRecalc!=null)
		SDArray=nodesToRecalc.cloneArray();
		contrastsTask.calculateNumbers(theTree, states2, SDArray, null, CommandRecord.nonscriptingRecord);
		xArray.resetSize(numNodes);
		contrastsTask.setOption(ContrastForChar.NODEVALUE,CommandRecord.nonscriptingRecord);
		double nodeValues[]=categoricalCalculator.getNodeValues();
		for (int i=0;i<xArray.getSize();i++)
			xArray.setValue(i, nodeValues[i]);
		statesOfNodes=categoricalCalculator.getAllStates();
		//	contrastsTask.calculateNumbers(theTree, states1, xArray, null, CommandRecord.nonscriptingRecord);
		yArray.resetSize(numNodes);
		contrastsTask.calculateNumbers(theTree, states2, yArray, null, CommandRecord.nonscriptingRecord);
		rawHeightArray.resetSize(numNodes);
		contrastsTask.setOption(ContrastForChar.RAWHEIGHT,CommandRecord.nonscriptingRecord);
		contrastsTask.calculateNumbers(theTree, states2, rawHeightArray, null, CommandRecord.nonscriptingRecord);
		heightArray.resetSize(numNodes);
		contrastsTask.setOption(ContrastForChar.CORRECTEDHEIGHT,CommandRecord.nonscriptingRecord);
		contrastsTask.calculateNumbers(theTree, states2,heightArray, null, CommandRecord.nonscriptingRecord);
		VPrimeArray.resetSize(numNodes);
		contrastsTask.setOption(ContrastForChar.VPRIME,CommandRecord.nonscriptingRecord);
		contrastsTask.calculateNumbers(theTree, states2,VPrimeArray, null, CommandRecord.nonscriptingRecord);

		reportNode(theTree,outputBuffer,theTree.getRoot(),lineDelimiter,columnDelimiter);

		outputBuffer.append("\n");
		return outputBuffer.toString();
	}


	/*.................................................................................................................*/
	public String getName() {
		return "FIC (AC3)";
	}

	/*.................................................................................................................*/
	public String getAuthors() {
		return "Franois Gao , Jonathan Corcy, Michel Laurin (based on PDAP)";
	}

	/*.................................................................................................................*/
	public String getVersion() {
		return "1.08";
	}
	/*.................................................................................................................*/
	public boolean isPrerelease() {
		return false;
	}
	/*.................................................................................................................*/
	public boolean isSubstantive(){
		return true;
	}   	 

	/*.................................................................................................................*/
	/** returns an explanation of what the module does.*/
	public String getExplanation() {
		return "Generates FIC format report files as described in Garland et al. 1993" ;
	}

	/*.................................................................................................................*/
	/** returns current parameters, for logging etc..*/
	public String getParameters() {
		return "" ;
	}

}

