/* AC3 Mesquite module */

package mesquite.ac3.RootReconstructions;

/* ~~ */

import mesquite.lib.*;
import mesquite.lib.characters.*;
import mesquite.lib.duties.*;
import mesquite.ac3.lib.*;
import mesquite.categ.lib.CategoricalDistribution;
import mesquite.cont.lib.*;
import JSci.maths.statistics.*;

/* ======================================================================== */
public class RootReconstructions extends RootRegressionCalculator {
    Tree lastUsedTree = null;

    TreeReference lastTreeReference;

    String DiagnosticString;

    AC32CTStatPakSource sps;

    private AC32CTStatPak myStats = null;

    MesquiteBoolean verboseQueryMode;

    MesquiteChart chart;
    AC3TreeWinAsstC chartModule = null;

    /* ................................................................................................................. */
    public boolean startJob(String arguments, Object condition,
            CommandRecord commandRec, boolean hiredByName) {
        verboseQueryMode = new MesquiteBoolean(false);
        loadPreferences();
        // verboseQueryMode.setValue(false);
        lastTreeReference = null;
        addMenuItem("Root Calculation Query Mode...", makeCommand("setVerboseQueryMode", (Commandable) this));
        addMenuItem("Set CI/PI width...", makeCommand("setwidth",(Commandable) this));
        addMenuItem("Set Polytomy DF Reduction...", makeCommand("setDFReduction", (Commandable) this));
        sps = (AC32CTStatPakSource) hireNamedEmployee(commandRec,AC32CTStatPakSource.class, "#SPRootReconstruction");
        return true;
    }

    /*-----------------------------------------*/
    public void processPreferencesFromFile(String[] prefs) {
        if (prefs != null && prefs.length > 0) {
            verboseQueryMode.setValue("v".equals(prefs[0]));
        }
    }

    /*-----------------------------------------*/
    public String[] preparePreferencesForFile() {
        String[] prefs = new String[1];
        if (verboseQueryMode.getValue())
            prefs[0] = "v";
        else
            prefs[0] = "n";
        return prefs;
    }

    /*-----------------------------------------*/
    public Snapshot getSnapshot(MesquiteFile file) {
        Snapshot temp = new Snapshot();
        temp.addLine("setwidth1 " + myStats.getWidth1());
        temp.addLine("setwidth2 " + myStats.getWidth2());
        temp.addLine("setDFReduction " + myStats.getDFReduction());
        return temp;
    }

    /*-----------------------------------------*/
    /**  */
    public Object doCommand(String commandName, String arguments,
            CommandRecord commandRec, CommandChecker checker) {
        if (checker.compare(this.getClass(),"Sets the first width for the confidence and prediction intervals","[a confidence width e.g., 0.95]", commandName,"setwidth1")) {
            if (myStats == null)
                return null;
            MesquiteInteger pos = new MesquiteInteger(0);
            double newWidth = MesquiteDouble.fromString(arguments, pos);
            if (!MesquiteDouble.isCombinable(newWidth))
                myStats.querySetWidth();
            else
                myStats.setWidth1(newWidth);
            parametersChanged(commandRec);
        }
        else if (checker.compare(this.getClass(),"Sets the second width for the confidence and prediction intervals","[a confidence width e.g., 0.90]", commandName,"setwidth2")) {
            if (myStats == null)
                return null;
            MesquiteInteger pos = new MesquiteInteger(0);
            double newWidth = MesquiteDouble.fromString(arguments, pos);
            if (!MesquiteDouble.isCombinable(newWidth))
                myStats.querySetWidth();
            else
                myStats.setWidth2(newWidth);
            parametersChanged(commandRec);
        } 
        else if (checker.compare(this.getClass(),"Sets both widths for the confidence and prediction intervals","[two confidence widths e.g., 0.95 0.90]", commandName,"setwidth")) {
            if (myStats == null)
                return null;
            MesquiteInteger pos = new MesquiteInteger(0);
            double newWidth1 = MesquiteDouble.fromString(arguments, pos);
            MesquiteInteger pos2 = null;
            double newWidth2 = MesquiteDouble.unassigned;
            if (MesquiteDouble.isCombinable(newWidth1)) {
                pos2 = new MesquiteInteger(arguments.indexOf(' ', 0));
                if (pos2.getValue() > 0)
                    newWidth2 = MesquiteDouble.fromString(arguments, pos2);
            }
            if (!MesquiteDouble.isCombinable(newWidth1)
                    || !MesquiteDouble.isCombinable(newWidth2))
                myStats.querySetWidth();
            else {
                myStats.setWidth1(newWidth1);
                myStats.setWidth2(newWidth2);
            }
            parametersChanged(commandRec);
        } 
        else if (checker.compare(this.getClass(),"Sets the number of DF's reduced because of polytomies","[number of dof]", commandName, "setDFReduction")) {
            if (myStats == null)
                return null;
            MesquiteInteger pos = new MesquiteInteger(0);
            int dfReduction = MesquiteInteger.fromFirstToken(arguments, pos);
            if (!MesquiteInteger.isCombinable(dfReduction))
                myStats.queryDFReduce(lastUsedTree, true);
            else
                myStats.setDFReduction(dfReduction);
            parametersChanged(commandRec);
        } 
        else if (checker.compare(this.getClass(),"Sets the query mode to verbose or not", "[on/off]",commandName, "setVerboseQueryMode")) {
            if (StringUtil.blank(arguments)) {
                MesquiteBoolean ok = MesquiteBoolean.queryCheckBox(containerOfModule(),"Set Verbose Query Preference","In verbose mode you will be asked for settings for AC3 root reconstruction frequently; otherwise you change the settings using menu items","Use Verbose Mode", verboseQueryMode.getValue());
                if (!ok.isUnassigned()) {
                    verboseQueryMode.setValue(ok.getValue());
                    storePreferences();
                }
            } else {
                verboseQueryMode.toggleValue(parser.getFirstToken(arguments));
                storePreferences();
            }
        } else
            return super.doCommand(commandName, arguments, commandRec, checker);
        return null;
    }

    /*-----------------------------------------*/
    public void setChartModule(AC3TreeWinAsstC module) {
        chartModule = module;
        myStats = sps.getStatPak(chartModule.getDFReduction());
    }
    
    /**
     * Simply returns the statPak so the charter can access the slope and other
     * interesting results.
     */
    public AC32CTStatPak getStatPak() {
        if (chartModule == null){
            MesquiteMessage.warnProgrammer("No chartModule set for "+this);
            return null;
        }
        if (myStats == null){
            myStats = sps.getStatPak(chartModule.getDFReduction());
        }
        return myStats;
    }

    /*-----------------------------------------*/
    public void setChart(MesquiteChart chart) {
        this.chart = chart;
    }

    /*-----------------------------------------*/
    /** This is a nop for now - maybe remove this from the abstract class?? */
    public void setOption(int option) {
    }

    /* ................................................................................................................. */
    public void calculateNumbers(Tree tree,
            CharacterDistribution charDistribution1,
            CharacterDistribution charDistribution2, NumberArray result,
            MesquiteString resultString, CommandRecord commandRec) {
        CategoricalDistribution observedStates1 = null;
        ContinuousDistribution observedStates2 = null;
        MesquiteDouble conSlope = new MesquiteDouble();
        int root; // don't initialize until we know tree is good
        Tree reducedTree;

        if (result == null || tree == null)
            return;
        if (!(charDistribution1 instanceof CategoricalDistribution)
                || !(charDistribution2 instanceof ContinuousDistribution))
            return;
        result.zeroArray();
        root = tree.getRoot();
        observedStates1 = (CategoricalDistribution) charDistribution1;
        observedStates2 = (ContinuousDistribution) charDistribution2;
        //      Peter: tighter restrictions; any change will provoke query or reset
        if ((lastTreeReference == null) ||
            (tree instanceof MesquiteTree && 
             !((MesquiteTree) tree).sameTreeVersions(lastTreeReference, true, true))) { 

            // first create and initialize the new rootStatPak
            myStats = sps.getStatPak(chartModule.getDFReduction());

            // 
            reducedTree = tree;
            // now set the tree so doCalculations will do the right thing...
            myStats.setTree(tree);
            myStats.countInternalNodes(tree, root); // count the internal nodes,
                                                    // to set number
            int zlb = myStats.countZLBs(tree, root); // Zero-Length-Branches are what
                                            // AC3 calls "polytomies"; count to
                                            // inform the user
            // now ask the user to set df Reductions
            if (!commandRec.scripting()){
                if (!chartModule.getDFQueryDone() && zlb > 0){
                    myStats.queryDFReduce(tree,true);
                    chartModule.setDFQueryDone(true);
                }
                else if (verboseQueryMode.getValue()){
                    myStats.queryDFReduce(tree, true);
                    chartModule.setDFQueryDone(true);
                    myStats.querySetWidth();
                }
            }
        }
        lastUsedTree = tree;
        lastTreeReference = ((MesquiteTree) tree)
                .getTreeReference(lastTreeReference);
        if ((observedStates1 != null) || (observedStates2 != null)) {
            NumberArray xArray = new NumberArray();
            NumberArray yArray = new NumberArray();
            myStats.setObserved1(observedStates1);
            myStats.setObserved2(observedStates2);
            boolean munch = myStats.doCalculations(xArray, yArray, chart,
                    commandRec);
            // Again, results is never set.
            if (munch)
                chart.munch(commandRec);
            if (resultString != null)
                resultString.setValue("Reconstructed Root Values");

        }

    }

    /* ................................................................................................................. */

    /*.................................................................................................................*/
    public Class getCharacterClass() {
        return ContinuousState.class;
    }

    /*.................................................................................................................*/
    public CompatibilityTest getCompatibilityTest() {
        return new ContinuousStateTest();
    }

    /** returns current parameters, for logging etc..*/
    /*.................................................................................................................*/
    public String getParameters() {
        return "Reconstructed root stats";
    }

    /*.................................................................................................................*/
    public String getAuthors() {
        return "Franois Gao , Jonathan Corcy, Michel Laurin (based on PDAP)";
    }

    /*.................................................................................................................*/
    public String getVersion() {
        return "1.08";
    }

    /*.................................................................................................................*/
    public boolean isPrerelease() {
        return false;
    }

    /*.................................................................................................................*/
    public String getName() {
        return "Reconstructed Root Value for AC3";
    }

    /*.................................................................................................................*/
    public String getExplanation() {
        return "Calculates statistics related to a reconstruction of the character at the root.";
    }


}
