/* AC3 Mesquite module */

// This class handles the calculations of root confidence and prediction intervals and the confidence
// and prediction intervals around the regression line run through the root projected back into the
// tip data plot (screens 9A and 9B).

package mesquite.ac3.SPRootReconstruction;

import java.awt.*;
import java.util.*;
import mesquite.lib.*;
import mesquite.lib.characters.CharacterState;
import mesquite.ac3.lib.AC3ContrastCalculator;
//import mesquite.ac3.lib.BivariateContrastCalculator;
import mesquite.ac3.lib.NRtDistribution;
import mesquite.ac3.lib.NoPolyTree;
import mesquite.ac3.lib.AC32CTStatPak;
import mesquite.ac3.lib.AC32CTStatPakSource;
//import mesquite.ac3.lib.AC3StatPak;
import mesquite.ac3.lib.UnivariateContrastCalculator;
import mesquite.categ.lib.CategoricalDistribution;
import mesquite.cont.lib.*;
//import JSci.maths.statistics.*;    // for p-value calculations

/** */


public class SPRootReconstruction extends AC32CTStatPakSource  {
	public boolean startJob(String arguments, Object condition, CommandRecord commandRec, boolean hiredByName) {
		return true;
	}

	public AC32CTStatPak getStatPak(MesquiteInteger dfr){
		return new RootStatPak(dfr);
	}
	
	public String getName(){
		return "Source of statPaks for root reconstruction for AC3";
	}
		
    /*.................................................................................................................*/
    public String getAuthors() {
        return "Franois Gao , Jonathan Corcy, Michel Laurin (based on PDAP)" ;
    }
    /*.................................................................................................................*/
	/** returns module version */
    public String getVersion() {
        return "1.08";
    }
	/*.................................................................................................................*/
  	 public boolean isPrerelease() {
		return false;
   	}
	/*.................................................................................................................*/
   	public boolean isSubstantive() {
   		return false;  // the statPak is substantive, but I don't think this module is...
   	}
	public String getExplanation() {
		return "Calculates summary statistics and regression coefficients for a scatterplot.";
	}


}




class RootStatPak extends AC32CTStatPak{

	// These are options that control the display behavior (but not the calculating behavior) of
	// this statPak.

         
	public RootStatPak(MesquiteInteger dfr) {
		super(dfr);
	}


	/*-----------------------------------------*/
	/** set the drawing option to either draw the crosshairs or the regression lines.
	@param option - the choice for drawing option - either ROOTCI (draw crosshairs) or
	REGRESSIONCIPI (draw regression line and confidence and prediction interval curves
	around the regression. 
	Note that a bad value passed here will only cause the statpak to not draw anything,
	not crash. */
	public void setDrawingOption(int option) {
		drawingOption = option;
	}	
	
	/*-----------------------------------------*/
	/** Return the value of the first CI/PI width, which is also used in the
	plots in CI/PI plot (screen 9B) */
	public double getWidth1(){
		return width1;
	}
	
	/*-----------------------------------------*/
    /** Since the CI/PI widths are never actually used outside the StatPak, the
    method to allow the user to set them is here too. 
    @param w - the value assigned to the CI/PI width, which is also used in the 
    plots in screens 9A and 9B. */
	public void setWidth1(double w){
			width1 = w;			
	}

	/*-----------------------------------------*/
	/** 
	@return double - value of the second CI/PI width, which is also used in the
	plots in CI/PI plot (screen 9B) */
	public double getWidth2(){
		return width2;
	}
	
	/*-----------------------------------------*/
    /** @param w - the value assigned to the second CI/PI width, also used in plots for
    screens 9A and 9B. */
	public void setWidth2(double w){
		width2 = w;			
	}

	/*-----------------------------------------*/
	/** This prompts the user and sets the CI/PI widths. */
	public void querySetWidth(){
        MesquiteDouble tmp1 = new MesquiteDouble(width1);
        MesquiteDouble tmp2 = new MesquiteDouble(width2);
        MesquiteBoolean ans = new MesquiteBoolean();
                
        MesquiteDouble.queryTwoDoubles(null,
		                               "Interval Widths",
		                               "Enter the first (2-tailed) p-value width for CI/PI plots",
		                               "Enter the second (2-tailed) p-value width for CI/PI plots",
		                               ans,
		                               tmp1,
		                               tmp2);
		if (tmp1.isCombinable()) {
		    double tp1 = tmp1.getValue();
		    if ((tp1 <= 1.0) && (tp1 >= 0.0))
		       width1 = tp1;
		}
		if (tmp2.isCombinable()){
		     double tp2 = tmp2.getValue();
		     if ((tp2 <= 1.0) && (tp2 >= 0.0))
		         width2 = tp2;
		}
	}

	/** This prompts the user and sets Vh */  	
  	public void querySetVh(){
  			double tmp;
  			String contempString;

  			setVhToMeanHeight(tree);
  			if (Vh != MesquiteDouble.unassigned) {
  			    if (getContempTips())
  			  	    contempString = "All tips are contemporaneous in the uncorrected tree.";
  			    else
  				    contempString = "The tips are not contemporaneous in the uncorrected tree.";
  			    tmp = MesquiteDouble.queryDouble(null,
  			                                    "Vh",
  				  			      			    "The calculations for prediction intervals effectively assume " +
  							 			        "that the species to be predicted is attached at the root " +
  							 			        "(basal node) of the phylogenetic tree (see Garland and Ives, " +
  							 			        "2000, Am. Nat. 155:346-364).  By default, the branch length (Vh) " +
  							 			        "leading to the predicted species is assumed to equal the average\n" +
  							 			        "length\n of the (uncorrected) branches from root to tips. \n\n" +
  			                                     contempString + "\n\n Enter a value for Vh\n",       
  			                                 	Vh,
  			                                    0,
  			                                 	MesquiteDouble.infinite);
  			    if (MesquiteDouble.isCombinable(tmp))
  			        setVh(tmp);
  			}
  			else AlertDialog.notice(null, "Vh undefined",
  			  					   "Vh could not be estimated and is not meaningful because one or more " +
  			  					   "branches in the tree have undefined branch lengths.  Either assign lengths " +
  			  					   "or use a constant transform to set all branches equal to one.  No confidence or " +
  			  					   "prediction intervals will be displayed.");
  	}	
	

	/* The following are a set of accessors for slots that will be filled after a call to doCalculations - shouldn't be any reason to set them directly */
	
	/** returns the confidence interval around the reconstructed root value */
	public double getRootCI(int which) {
		if (which >= 0 && which < 2)
			return ci[which];
		return MesquiteDouble.unassigned;
	}		

	/** returns the standard error of the reconstructed root value */
	public double getRootSE(int which) {
		if (which >= 0 && which < 2)
			return se[which];
		return MesquiteDouble.unassigned;
	}
	
	/** returns the sum of the regression residuals */
	public double getResidualSum() {
		return sumresid;
	}
	
	/** returns the confidence interval around the regression slope */
	public double getSlopeCI() {
		return slopeCI;
	}
	
	/** returns the confidence interval around the regression y-intercept */
	public double getInterceptCI() {
		return interCI;
	}

	/** returns the standard error of the regression slope */
	public double getSlopeSE() {
		return slopeSE;
	}
	
	/** returns the standard error of the regression y-intercept */
	public double getInterceptSE() {
		return interSE;
	}

	/** returns the sum of squares for x */
	public double getX2Sum() {
		return x2Sum;
	}
	
	/** returns the  sigma-squared y value see Garland Midford and Ives 1999 */
	public double getSigmaSquaredY() {
		return sigma2y;
	}
	
	/** This allows a new value to be added to the vector of x values that will be added to the end of the
	       CI/PI table that appears in the text pane of screen 9B. */       
	public void addCIPIValue(MesquiteDouble x) {
	     CIPIValues.addElement(x);
    }
        
    /** This allows clearing the list of values (like if a character is changed) */
    public void clearCIPIlist() {
        CIPIValues = new Vector();
    }
        
    public MesquiteDouble CIPIValueAt (int i){
    	return (MesquiteDouble) CIPIValues.elementAt(i);
    }
        
	/*-----------------------------------------*/
	/** This method sets the maximum and minimum tip values in the appropriate fields of
	the statPak */
	private void calcBounds(Tree tree, double [] data0, double [] data1, int node) {
		if (tree.nodeIsInternal(node))
		 	for (int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter)) {
				     calcBounds(tree, data0, data1, daughter);
		  	}		
		else if (MesquiteDouble.isCombinable(data0[node]) &&
			   MesquiteDouble.isCombinable(data1[node])) {
			if (data0[node] > max[0])	
				max[0] = data0[node];
			if (data1[node] > max[1])
				max[1] = data1[node];
			if (data0[node] < min[0])
				min[0] = data0[node];
			if (data1[node] < min[1])
				min[1] = data1[node];
		}
	}	
	
	/*-----------------------------------------*/
	/**  A simple method to calculate the height of a tip, used to calculate the mean tip height as a proposed default for Vh */
	private double quickHeight(Tree tree, int node) {
		int tmp = node;
		double height = 0.0;
		do {
			height += tree.getBranchLength(tmp);
			tmp = tree.motherOfNode(tmp);
		}
		while(tmp != tree.getRoot()) ;
		return height;
	}

	/*-----------------------------------------*/
   	/** This function calculates the mean height of all the tree's tips.  
   	This provides a default value for Vh.   Vh is the assumed height 
   	of a tip when calculating a prediction interval.  The program reports
        whether tips are contemporaneous, using a tolerance of 1E-10, the 
        same tolerance used for deciding whether a tip has a depth of zero. */                   
	public void setVhToMeanHeight(Tree tree){
		int tipList [] = tree.getTerminalTaxa(tree.getRoot());
		double oldHeight = quickHeight(tree,tree.nodeOfTaxonNumber(tipList[0]));
		double sum = oldHeight;
		double currentHeight;
		boolean assignedBranches = true;
		
		for (int i=1;i< tipList.length;i++)  // already did tip 0.
			tipList[i] = tree.nodeOfTaxonNumber(tipList[i]);
		contempTips = true;
		for(int i = 1;(i<tipList.length) && assignedBranches;i++) {
			currentHeight = quickHeight(tree, tipList[i]);
			if (MesquiteDouble.isCombinable(currentHeight)) {
			    sum += currentHeight;
			    if (Math.abs((oldHeight-currentHeight)/currentHeight) > 1E-10) 
			        contempTips = false;
			    oldHeight = currentHeight;
			}
			else assignedBranches = false;
		}
		if (assignedBranches) 
		    Vh = sum/tipList.length;
		else 
		    Vh = MesquiteDouble.unassigned;
	}
	
	/*-----------------------------------------*/
	/** This just returns the contempTips flag, which holds the
	answer to the question whether all tips have the same height from
	the root. */
	public boolean getContempTips(){
		return contempTips;
	}
	
	/*-----------------------------------------*/
	public double getPValue(){
		return MesquiteDouble.unassigned;
	}
	/*-----------------------------------------*/
	
	/** This returns the appropriate mean, which is the reconstructed root value. */
	public void calcMeans(Tree tree, double [] data1, double [] data2) {
		mean[0] = data1[tree.getRoot()];
		mean[1] = data2[tree.getRoot()];
	}

	/*-----------------------------------------*/
	/** This calculates variances of contrasts (and cross products for later correlation
	 calculations). */
	public void calcVars(Tree tree, double[] contrasts1, double contrasts2 []) {
		variance[0] = 0.0;
		variance[1] = 0.0;
		varianceSums(tree,contrasts1,contrasts2,tree.getRoot());
      		variance[0] /= getNumber();  			// rate of evolution of trait 1
      		variance[1] /= getNumber();  			// rate of evolution of trait 2
	}

	/** This calculates the sums for getVars */
	private void varianceSums(Tree tree, double[] contrasts1, double[] contrasts2, int node) {
     	for (int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter))
			varianceSums(tree, contrasts1, contrasts2, daughter);
	 	if (tree.nodeIsInternal(node) && 
	 	    MesquiteDouble.isCombinable(contrasts1[node]) &&
	 	    MesquiteDouble.isCombinable(contrasts2[node])) {
	 	     	number++;
			variance[0] += contrasts1[node]*contrasts1[node];
			variance[1] += contrasts2[node]*contrasts2[node];
			corr += contrasts1[node]*contrasts2[node];
		}
	}

	/*-----------------------------------------*/
	/** The "v" values for the root are calculated and cached here. */
	private void setVs(Tree tree, double[] treeVPrimes) {
		int root = tree.getRoot();
      	v1 = treeVPrimes[tree.firstDaughterOfNode(root)];
      	v2 = treeVPrimes[tree.nextSisterOfNode(tree.firstDaughterOfNode(root))];
		v_adjust = (v1*v2)/(v1+v2);
	}

	/*-----------------------------------------*/
	/** This calculates the sum of squared X's (redundant with a variance?) and the
	sum of squared residuals from the contrast regression (through the origin). */
	private void getResidualSums(Tree tree, 
	                             double [] contrasts1, 
	                             double [] contrasts2, 
	                             double b, 
	                             int node) {
     	for (int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter))
			getResidualSums(tree, contrasts1, contrasts2, b, daughter);
	 	if (tree.nodeIsInternal(node) &&
	 	    MesquiteDouble.isCombinable(contrasts1[node]) &&
	 	    MesquiteDouble.isCombinable(contrasts2[node])) {
        		x2Sum += contrasts1[node]*contrasts1[node];
            	double temp = contrasts2[node] - b*contrasts1[node];
            	sigma2y += temp*temp;
        	}
	}

	/** This method calculates sums of squares for data1, and cross products for the internal nodes
	 of the tree rooted at node.  In this module, this is only used to calculate regression information
	 for the contrasts.*/
	private void getCSums(Tree tree, 
						  double[] data1, 
						  double[] data2,
						  MesquiteDouble sx2,
						  MesquiteDouble sxy,
						  int node) {
		if (tree.nodeIsInternal(node)) {  // only interior nodes
			if (MesquiteDouble.isCombinable(data1[node]) && // with valid data
	 	     	    MesquiteDouble.isCombinable(data2[node])) { 
				sx2.setValue(sx2.getValue()+data1[node]*data1[node]);
				sxy.setValue(sxy.getValue()+data1[node]*data2[node]); 
			}
			getCSums(tree,data1,data2,sx2,sxy,tree.firstDaughterOfNode(node));
			getCSums(tree,data1,data2,sx2,sxy,tree.lastDaughterOfNode(node));
		}
	}
	
	/** This method returns the ordinary least squares slope of the contrasts 
	regression in the MesquiteDouble conSlope.  As usual for contrasts, 
	the regression is through the origin.  */
	private void contrastSlope(Tree tree,
	                           double [] contrasts1,
	                           double [] contrasts2,
				  		       MesquiteDouble conSlope,
				  		       int root) {
		MesquiteDouble sx2 = new MesquiteDouble(0.0);
		MesquiteDouble sxy = new MesquiteDouble(0.0);
				
		getCSums(tree,contrasts1,contrasts2,sx2,sxy,root);
      	if (sx2.getValue() != 0) 
        	conSlope.setValue(sxy.getValue()/sx2.getValue());
    }

	/*-----------------------------------------*/
	/** This method calculates the CI for the intercept.  Width specifies   
   	 the level of confidence (usually 0.90 or 0.95).  This code implements 
   	 equations 14 (for slope CI) and 25 (for yintercept CI) from the      
   	 appendix to Garland, T. Jr. and A. Ives, 2000.  Using the past to 
   	 predict the present: Confidence intervals for regression equations in 
   	 phylogenetic comparative methods. American Naturalist. */        
   	private void slopeInterCI(Tree tree, double [] contrasts1, double [] contrasts2, double ls) {
   	    double Xz;
        double rawCI= tStats.CI(width1); 
      	
		x2Sum = 0.0;
		sigma2y = 0.0;
		getResidualSums(tree, contrasts1, contrasts2, ls, tree.getRoot());
        if (x2Sum != 0.0) {
            sigma2y /= getNumber();        // Note that s.number = # of tips - 1
            slopeSE = Math.sqrt(sigma2y*getNumber()/(x2Sum*getNumberMinus1()));
            Xz = mean[n1];       // Xz only, since Xi = 0
            interSE = Math.sqrt((sigma2y*getNumber()/getNumberMinus1())*(((Xz*Xz)/x2Sum)+v_adjust));
            interCI = rawCI*interSE;
            slopeCI = rawCI*slopeSE;
        }
        else {
        	interCI = -99999;
         slopeCI = -99999;
        }
	}


	/*-----------------------------------------*/
	/** This method calculates the confidence intervals around the root estimates and
            calls slopeInterCI to get confidence intervals for the regression estimates. 
            Root C.I.'s are calculated using equation A9 of Garland, Midford, and Ives 1999. 
            Intercept C.I.'s are calculated according to equation A13 of Garland and Ives 2000.    */
	private void calcRootIntervals(Tree tree, 
						           double contrasts1[], 
						           double contrasts2[], 
						           MesquiteDouble conSlope) {
		// Note the difference in degrees of freedom between the root reconstruction
		// and the regression statistics.
		// creating a new distribution seems cleaner than setting and restoring the
		// df of the main tStats object. 						  
		NRtDistribution tStats2 = new NRtDistribution(getNumber()-getDFReduction());
		double rootWidth = tStats2.CI(width1);
		
      	se[0] = Math.sqrt(variance[0]*v_adjust);
      	se[1] = Math.sqrt(variance[1]*v_adjust);
		ci[0] = rootWidth*se[0];
		ci[1] = rootWidth*se[1];
		ls = conSlope.getValue();
		lsy = mean[1] - ls*mean[0];
		slopeInterCI(tree, contrasts1, contrasts2,ls);  
 	}

	/*-----------------------------------------*/
	/** This calculates reconstructed root values*/
	private void calcReconstructions(Tree tree, 
					  		double [] x1,
					  		double [] x2,
	                      	double [] contrasts1, 
	                      	double [] contrasts2, 
	                      	double [] treeVPrimes,
	                         MesquiteDouble conSlope,
	                         CommandRecord commandRec) {

		int root = tree.getRoot();
		
		n1 = 0;
		n2 = 1;
		calcBounds(tree,x1,x2,root);     // not the same as the getBounds for regular stats.
		calcMeans(tree,x1,x2);
		number = 0;
		calcVars(tree,contrasts1,contrasts2);
		setVs(tree,treeVPrimes);
		calcRootIntervals(tree,contrasts1,contrasts2,conSlope);    
	}

	/*.................................................................................................................*/
	private double[] contrasts1, contrasts2;
	private double[] x1,x2;
	private double[] vPrime1;
	private NumberArray	  contrastWithNodes;
	private boolean setRootChartLimits(double minX, double maxX, double minY, double maxY, MesquiteChart chart) {
		boolean result = false;
		MesquiteNumber setter;
		
		if (chart != null){    //otherwise nothing to work with...
			// checking for chart.isMinimumXConstrained won't work, doCalculations resets it now
			if (chart.getMinimumX()==null) {  //no constraint value, so just set it
				setter = new MesquiteNumber(minX);
				chart.constrainMinimumX(setter);
				result = true;
			}
			else {  // a constraint exists, so adjust it
				if (chart.getMinimumX().getDoubleValue() > minX) {
					double growth =  1.5*(chart.getMinimumX().getDoubleValue() -minX); // 1% should overwhelm any conversion problem
					setter = new MesquiteNumber(chart.getMinimumX().getDoubleValue()-growth);
					chart.constrainMinimumX(setter);  
					result = true;
				}
			}
			if (chart.getMaximumX()==null) {  //no constraint, so just set it
				setter =  new MesquiteNumber(maxX);
				chart.constrainMaximumX(setter);
				result = true;
			}
			else {  // a constraint exists, so adjust it
				if (chart.getMaximumX().getDoubleValue() < maxX) {
					double growth = 1.5*(chart.getMaximumX().getDoubleValue() - maxX); //1% should overwhelm any conversion problem
					setter = new MesquiteNumber(chart.getMaximumX().getDoubleValue() - growth);
					chart.constrainMaximumX(setter);  
					result = true;
				}
			}
			if (chart.getMinimumY()==null) {  //no constraint, so just set it
				setter = new MesquiteNumber(minY);
				chart.constrainMinimumY(setter);
				result = true;
			}
			else {  // a constraint exists, so adjust it			
				if (chart.getMinimumY().getDoubleValue() > minY) {
					double growth = 1.5*(chart.getMinimumY().getDoubleValue() -minY); // 1% should overwhelm any conversion problem
					setter = new MesquiteNumber(chart.getMinimumY().getDoubleValue()-growth);
					chart.constrainMinimumY(setter);  
					result = true;
				}
			}
			if (chart.getMaximumY()==null) {  //no constraint, so just set it
				setter = new MesquiteNumber(maxY);
				chart.constrainMaximumY(setter);
				result = true;
			}
			else {  // a constraint exists, so adjust it
				if (chart.getMaximumY().getDoubleValue() < maxY) {
					double growth = 1.5*(chart.getMaximumY().getDoubleValue() - maxY);
					setter =  new MesquiteNumber(chart.getMaximumY().getDoubleValue() - growth);
					chart.constrainMaximumY(setter);  // 1% should overwhelm any conversion foo
					result = true;
				}
			}
		}
		return result;		
	}

	private boolean setRegressionChartLimits(double minY, double maxY, MesquiteChart chart) {
		boolean result = false;
		MesquiteNumber setter;
		if (chart != null){    //otherwise nothing to work with...
			if (chart.getMinimumY()==null) {  //no constraint, so just set it
				setter = new MesquiteNumber(minY);
				chart.constrainMinimumY(setter);
				result = true;
			}
			else {  // a constraint exists, so adjust it			
				if (chart.getMinimumY().getDoubleValue() > minY) {
					double growth = 1.5*(chart.getMinimumY().getDoubleValue() -minY); // 1% should overwhelm any conversion problem
					setter = new MesquiteNumber(chart.getMinimumY().getDoubleValue()-growth);
					chart.constrainMinimumY(setter);  
					result = true;
				}
			}
			if (chart.getMaximumY()==null) {  //no constraint, so just set it
				setter = new MesquiteNumber(maxY);
				chart.constrainMaximumY(setter);
				result = true;
			}
			else {  // a constraint exists, so adjust it
				if (chart.getMaximumY().getDoubleValue() < maxY) {
					double growth = 1.5*(chart.getMaximumY().getDoubleValue() - maxY);
					setter = new MesquiteNumber(chart.getMaximumY().getDoubleValue() - growth);
					chart.constrainMaximumY(setter);  // 1% should overwhelm any conversion foo
					result = true;
				}
			}
		}
		return result;		
	}
		
	/** Note: in order to get this method to work with these arguments, the RootStatPak needs to have
	had tree and obs1 & obs2 previously set.  After this runs, all the values will be in place for a
	root reconstruction plot or a CI/PI plot, though if Vh is not set, the Vh=0 default makes the CI
	and PI widths identical. 
	@return boolean: true if the chart needs to remunch.*/
	
	public boolean doCalculations(NumberArray xArray, NumberArray yArray, MesquiteChart chart, CommandRecord commandRec){

	/**	MesquiteDouble conSlope = new MesquiteDouble();
		BivariateContrastCalculator calculator;
		boolean xOk = false;
		boolean result = false;
		min[0] = MesquiteDouble.infinite;
		min[1] = MesquiteDouble.infinite;
	 	max[0] = MesquiteDouble.negInfinite;
	 	max[1] = MesquiteDouble.negInfinite;
		
		resetChartBounds(chart);    // this seems to be necessary for the charts to scale properly
		
		calculator = new BivariateContrastCalculator(tree, obs1, obs2);
		contrasts1 = calculator.contrastCalculation();
		xOk = calculator.Ok();
		contrasts2 = calculator.getContrasts2();
		x1 = calculator.getNodeValues();
		x2 = calculator.getNodeValues2();
		vPrime1 = calculator.getVPrime();
		tStats = new NRtDistribution(getNumberMinus1()-getDFReduction());
		contrastSlope(tree,contrasts1,contrasts2,conSlope,root);
		calcReconstructions(tree,x1,x2,contrasts1,contrasts2,vPrime1,conSlope,commandRec);*/

		UnivariateContrastCalculator continuousCalculator;
		AC3ContrastCalculator categoricalCalculator;
		MesquiteDouble conSlope = new MesquiteDouble();
		boolean xOk = false;
		boolean result = false;
		min[0] = MesquiteDouble.infinite;
		min[1] = MesquiteDouble.infinite;
	 	max[0] = MesquiteDouble.negInfinite;
	 	max[1] = MesquiteDouble.negInfinite;
	 	contrastWithNodes=null;
		
		resetChartBounds(chart);    // this seems to be necessary for the charts to scale properly
		categoricalCalculator= new AC3ContrastCalculator(tree,obs1);
		continuousCalculator = new UnivariateContrastCalculator(tree,obs2);
		contrasts1 = categoricalCalculator.contrastCalculation();
		contrasts2 = continuousCalculator.contrastCalculation();
		if(categoricalCalculator.getNodesToRecalculate()!=null){
			contrastWithNodes=categoricalCalculator.getNodesToRecalculate().cloneArray();
		double[] res=continuousCalculator.contrastToRecalculate(contrastWithNodes);
		for (int j=0;j<res.length;j+=2){
			contrasts2[(int)res[j]]=(double)res[j+1];}
		}
		xOk = continuousCalculator.Ok();
		x1 = categoricalCalculator.getNodeValues();
		CharacterState statesOfRoot=obs1.getCharacterState(obs1.getCharacterState(), 2);
		String rootStates=statesOfRoot.toString();
		char[] rootStatesTable=rootStates.toCharArray();
		if(rootStatesTable.length==3){
		String aChar = new Character((char)rootStatesTable[1]).toString();
		x1[2]=Double.parseDouble(aChar);
		}
		else{
			double temp=0;
			double num=0;
		for(int i=1;i<rootStatesTable.length;i+=2){
			String aChar = new Character((char)rootStatesTable[i]).toString();
			temp+=Double.parseDouble(aChar);
			num++;
		}
		if(num!=0){
		temp/=num;
		x1[2]=temp;
		}}
		x2 = continuousCalculator.getNodeValues();
		vPrime1 = continuousCalculator.getVPrime();
		tStats = new NRtDistribution(getNumberMinus1()-getDFReduction());
		contrastSlope(tree,contrasts1,contrasts2,conSlope,root);
		calcReconstructions(tree,x1,x2,contrasts1,contrasts2,vPrime1,conSlope,commandRec);
		//overrides the ci intervals (only  for categorical characters)
		String tempCI= new Character((char)rootStatesTable[1]).toString();
		ci[0]=mean[0]-Double.parseDouble(tempCI);

		// finally, set some minimum and maximum limits for the confidence interval plot (8-20-02 PEM)
 		// calculate some maximum and minimum y values, so PI lines and the ends of the cross hairs fit on the display
		switch (drawingOption) {
			case ROOTCI: { 
				double minX = mean[0]-ci[0];
				double maxX = mean[0]+ci[0];
				double minY = mean[1]-ci[1];
				double maxY = mean[1]+ci[1];
				result = setRootChartLimits(minX,maxX,minY,maxY,chart);
				break;
			}
			case REGRESSIONCIPI: {
				if (MesquiteDouble.isCombinable(Vh)) {
				    double minCIPIY1 = ls*min[0]+lsy-calcPI(min[0],0,width1);
				    double maxCIPIY1 = ls*max[0]+lsy+calcPI(max[0],0,width1);
				    double minCIPIY2 = ls*min[0]+lsy-calcPI(min[0],0,width2);
				    double maxCIPIY2 = ls*max[0]+lsy+calcPI(max[0],0,width2);
				    result = setRegressionChartLimits(MesquiteDouble.minimum(minCIPIY1,minCIPIY2),
				                                                         MesquiteDouble.maximum(maxCIPIY1,maxCIPIY2),
				                                                         chart);
				}
				else result = false;
				break;
			}
		}
		return result;
	}

	/*-----------------------------------------*/
	/** This method calculates the confidence interval of the regression estimate about a single
	point.  This is used to the confidence interval lines and to generate the .CI file. */
	private double calcCI(double Xi,int index1,double width) {
        	double SEy;
         	double rawCI;
         	double deltaX = Xi-mean[index1];

         	SEy = Math.sqrt(sigma2y*(((deltaX*deltaX)/x2Sum)+v_adjust));
         	rawCI = tStats.CI(1-(1-width)/2);
         	return rawCI*SEy;
	}
	
	/*-----------------------------------------*/
	/** This method calculates the prediction interval of the regression estimate about a single
	point.  This is used to generate the confidence interval lines and to generate the .CI file. */
	private double calcPI(double Xi,int index1, double width) {
        	double SEyNew;
         	double rawCI;
         	double deltaX = Xi-mean[index1];

         	SEyNew = Math.sqrt(sigma2y*(Vh+((deltaX*deltaX)/x2Sum)+v_adjust));
         	rawCI = tStats.CI(1-(1-width)/2);
         	return rawCI*SEyNew;
  	}


    /** This handles the formating problem that MesquiteDouble.toFixedWidthString almost, but doesn't
        quite handle correctly. */
    private String formatForCIPI(double value,String tableDelimiter) {
        String tmp = MesquiteDouble.toFixedWidthString(value,15,false);
        if (value <0)
            tmp += " " + tableDelimiter;
        else 
            tmp += "  " + tableDelimiter;
        return tmp;
    }   


	// CI/PI file columns
	
    //   1.  Tip Name
    //   2.  X Value
    //   3.  Observed Y Value
    //   4.  Predicted Y Value (Yhat)
    //   5.  Lower ',w1:2,'% Confidence Interval');
    //   6.  Upper ',w1:2,'% Confidence Interval');
    //   7.  Lower ',w1:2,'% Prediction Interval');
    //   8.  Upper ',w1:2,'% Prediction Interval');
    //   9.  Lower ',w2:2,'% Confidence Interval');
    // 10.  Upper ',w2:2,'% Confidence Interval');
    // 11.  Lower ',w2:2,'% Prediction Interval');
    // 12.  Upper ',w2:2,'% Prediction Interval');

   /** This writes one line of the CIPI table */
    private void CIPILine(double x, String yString, String label, String lineDelimiter, String tableDelimiter, StringBuffer result){
	    double yhat = ls*x + lsy;
	    double CIWidth1 = calcCI(x,0,width1);
	    double PIWidth1 = calcPI(x,0,width1);
	    double CIWidth2 = calcCI(x,0,width2);
	    double PIWidth2 = calcPI(x,0,width2);
	    result.append(label + tableDelimiter + 
	 	              formatForCIPI(x,tableDelimiter) +  
	 	              yString +   
	 	              formatForCIPI(yhat,tableDelimiter) + 
	 	              formatForCIPI(yhat-CIWidth1,tableDelimiter) + 
	 	              formatForCIPI(yhat+CIWidth1,tableDelimiter) + 
	 	              formatForCIPI(yhat - PIWidth1,tableDelimiter) +
	 	              formatForCIPI(yhat + PIWidth1,tableDelimiter) +
	 	              formatForCIPI(yhat-CIWidth2,tableDelimiter) + 
	 	              formatForCIPI(yhat+CIWidth2,tableDelimiter) +
	 	              formatForCIPI(yhat - PIWidth2,tableDelimiter) +
	 	              formatForCIPI(yhat + PIWidth2,tableDelimiter) +
	 	              lineDelimiter);
    }
        
    private void CIPIBadPoint(double x, double y, String label, String lineDelimiter, String tableDelimiter, StringBuffer result) {
        String xStr;
        String yStr;
        String yhatStr;
            
        if (MesquiteDouble.isCombinable(x)) {
            xStr = formatForCIPI(x,tableDelimiter);
            yhatStr = formatForCIPI(ls*x+lsy,tableDelimiter);
        }
        else {
            xStr = "?                " + tableDelimiter;
            yhatStr = "?                " + tableDelimiter;
        } 
        if(MesquiteDouble.isCombinable(y))
            yStr = formatForCIPI(y,tableDelimiter);
        else
            yStr = "?                " + tableDelimiter;
            result.append(label + tableDelimiter + xStr + yStr + yhatStr + " No CI/PI values calculated for missing data" + lineDelimiter);
    }

	/** This generates the rows for CI/PI values for nodes in the tree*/
	private void CIPITreevalues(NoPolyTree tree, 
	                            CategoricalDistribution obs1, 
	                            ContinuousDistribution observed2, 
	                            int node, 
	                            String lineDelimiter, 
	                            String tableDelimiter,
                                boolean convertSpaces,
	                            StringBuffer result) {
	 	if (!tree.nodeIsInternal(node)) {
	 	   double x = obs1.getState(tree.taxonNumberOfNode(node));
	 	   double y = observed2.getState(tree.taxonNumberOfNode(node), 0);
           String label;
           if (convertSpaces)
               label = spaceConvertString(tree.getNodeLabel(node));
           else
               label = tree.getNodeLabel(node);
           if (MesquiteDouble.isCombinable(x) & MesquiteDouble.isCombinable(y))
	 	      CIPILine(x,formatForCIPI(y,tableDelimiter),label,lineDelimiter,tableDelimiter,result);
	 	   else 
	 	      CIPIBadPoint(x, y,label,lineDelimiter,tableDelimiter,result);
	    }
		else
        	for (int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter))
		        CIPITreevalues(tree, obs1, observed2, daughter,lineDelimiter,tableDelimiter,convertSpaces,result);
	}

    /** This is the main method for generating the CIPI table. */
	private void CIPIvalues(NoPolyTree tree, CategoricalDistribution obs1, ContinuousDistribution observed2, int rt,StringBuffer result,String lineDelimiter, String tableDelimiter,boolean convertSpaces) {
        CIPITreevalues(tree, obs1,observed2,rt,lineDelimiter,tableDelimiter,convertSpaces,result);
        for(int i = -10; i <= -1; i++){
        	double x = min[0] + min[0]*i/100.0;
	 	   	String iStr = MesquiteInteger.toString(i);  // random column straightening stuff
	 	   	if ((i <0) && (i > -10))  
	 	        iStr += " ";   
		   	CIPILine(x,"-99999            ",iStr,lineDelimiter,tableDelimiter,result);
	    } 
	    for(int i = 1; i <= 10; i++){
            double x = max[0]+max[0]*i/100.0;
	 	    String iStr = MesquiteInteger.toString(i);  // random column straightening stuff
	 	    if ((i >= 0) && (i < 10)) 
	 	        iStr += "  ";
	 	    else if (i == 10)
	 	       iStr += " ";   
		    CIPILine(x,"-99999            ",iStr,lineDelimiter,tableDelimiter,result);
	    } 
	    CIPILine(0.0,"-99999            ","00 ",lineDelimiter,tableDelimiter,result);
        for (int i=0; i < CIPIValues.size(); i++)
	       CIPILine(CIPIValueAt(i).getValue(),"-99999            ","-99",lineDelimiter,tableDelimiter,result);
    }

	/*-----------------------------------------*/
	/** This method returns a string describing the contents of the statPak.  This string is
	displayed at the bottom of the text pane of the diagnostic screen. */
 	public String flst() {

		if (drawingOption == ROOTCI) {
			StringBuffer results = new StringBuffer(400);
		
	        results.append("Number of tips:  " + (getNumber()+1) + "\n");
	        results.append("Number of independent contrasts:  " + getNumber() + "\n");
	        results.append("Number of counted polytomies:  " + polytomies + "\n");
	       	results.append("User-specified degrees of freedom reduction for computing C.I.'s:  " + getDFReduction() + "\n");
	       	results.append("\nList of recalculated nodes: ");
	       	if(contrastWithNodes!=null){
	 			for (int i=0;i<contrastWithNodes.getSize();i+=3){
	 				if (contrastWithNodes.getInt(i)!=0){
	 					boolean leftNodeIsTerm=tree.nodeIsTerminal(contrastWithNodes.getInt(i+1));
	 					boolean rightNodeIsTerm=tree.nodeIsTerminal(contrastWithNodes.getInt(i+2));
	 					if(leftNodeIsTerm && rightNodeIsTerm)
	 						results.append("\n\tTerminal taxa: "+ tree.getNodeLabel(contrastWithNodes.getInt(i+1))+"(Node "+ contrastWithNodes.getInt(i+1)+"), "+tree.getNodeLabel(contrastWithNodes.getInt(i+2)) +"(Node "+ contrastWithNodes.getInt(i+2)+") were contrasted at: Node "+contrastWithNodes.getInt(i));
	 					else if(leftNodeIsTerm && !rightNodeIsTerm)
	 						results.append("\n\tTerminal Taxon: "+ tree.getNodeLabel(contrastWithNodes.getInt(i+1)) +"(Node "+ contrastWithNodes.getInt(i+1)+") and internal node: Node "+ contrastWithNodes.getInt(i+2) +" were contrasted at: Node "+contrastWithNodes.getInt(i));
	 					else if(!leftNodeIsTerm && rightNodeIsTerm)
	 						results.append("\n\tTerminal Taxon: "+ tree.getNodeLabel(contrastWithNodes.getInt(i+2)) +"(Node "+ contrastWithNodes.getInt(i+2)+") and internal node: Node "+ contrastWithNodes.getInt(i+1) +" were contrasted at: Node "+contrastWithNodes.getInt(i));
	 					else
	 						results.append("\n\tInternal nodes: Node "+ contrastWithNodes.getInt(i+1) +", Node "+ contrastWithNodes.getInt(i+2) +" were contrasted at: Node "+contrastWithNodes.getInt(i));
	 					}
	 				else
	 					break;
	 			}
	 		}
	 		else results.append("None ");
	       	results.append("\n\nSlope (least squares):              " + ls + "\n");
	       	results.append("Standard Error:                     " + slopeSE + "\n");
	       	results.append(" Lower " + width1*100 + "% Confidence Interval:   " + (ls-slopeCI) + "\n");
	       	results.append(" Upper " + width1*100 + "% Confidence Interval:   " + (ls+slopeCI) + "\n");
	       	results.append("Intercept (least squares):         " + lsy + "\n");
	       	results.append(" Standard Error:                    " + interSE + "   / as mapped back onto \\ \n");
	       	results.append(" Lower " + width1*100 + "% Confidence Interval:  " + (lsy-interCI) + "  |  the original data    | \n");
	       	results.append(" Upper " + width1*100 + "% Confidence Interval:  " + (lsy+interCI) + "   \\ space               /  \n\n");
	       	results.append("                                            X = Trait " + (n1+1));
	       	results.append("       Y = Trait " + (n2+1) + "\n");
	       	results.append("Estimated Root Node Values:              " + mean[n1] + "\t  " + mean[n2] + "\n");
	       	results.append("Rates of Evolution:                      " + variance[n1] + " " + variance[n2] + "\n");
	       	results.append("Standard Errors of Root Node Estimates:  " + se[n1] + " " + se[n2] + "\n");
	       	results.append("Lower " + width1*100 + "% Confidence Intervals:        " + (mean[n1]-ci[n1]));
	       	results.append("\t" + (mean[n2]-ci[n2]) + "\n");
	       	results.append("Upper " + width1*100 + "% Confidence Intervals:        " + (mean[n1]+ci[n1]));
	       	results.append("\t" + (mean[n2]+ci[n2]) + "\n\n");
			return results.toString();
		}
		else  {// Return the CI/PI file inline...
		    StringBuffer results = new StringBuffer(1000);
		   results.append("List of recalculated nodes: ");
		   if(contrastWithNodes!=null){
	 			for (int i=0;i<contrastWithNodes.getSize();i+=3){
	 				if (contrastWithNodes.getInt(i)!=0){
	 					boolean leftNodeIsTerm=tree.nodeIsTerminal(contrastWithNodes.getInt(i+1));
	 					boolean rightNodeIsTerm=tree.nodeIsTerminal(contrastWithNodes.getInt(i+2));
	 					if(leftNodeIsTerm && rightNodeIsTerm)
	 						results.append("\n\tTerminal taxa: "+ tree.getNodeLabel(contrastWithNodes.getInt(i+1))+"(Node "+ contrastWithNodes.getInt(i+1)+"), "+tree.getNodeLabel(contrastWithNodes.getInt(i+2)) +"(Node "+ contrastWithNodes.getInt(i+2)+") were contrasted at: Node "+contrastWithNodes.getInt(i));
	 					else if(leftNodeIsTerm && !rightNodeIsTerm)
	 						results.append("\n\tTerminal Taxon: "+ tree.getNodeLabel(contrastWithNodes.getInt(i+1)) +"(Node "+ contrastWithNodes.getInt(i+1)+") and internal node: Node "+ contrastWithNodes.getInt(i+2) +" were contrasted at: Node "+contrastWithNodes.getInt(i));
	 					else if(!leftNodeIsTerm && rightNodeIsTerm)
	 						results.append("\n\tTerminal Taxon: "+ tree.getNodeLabel(contrastWithNodes.getInt(i+2)) +"(Node "+ contrastWithNodes.getInt(i+2)+") and internal node: Node "+ contrastWithNodes.getInt(i+1) +" were contrasted at: Node "+contrastWithNodes.getInt(i));
	 					else
	 						results.append("\n\tInternal nodes: Node "+ contrastWithNodes.getInt(i+1) +", Node "+ contrastWithNodes.getInt(i+2) +" were contrasted at: Node "+contrastWithNodes.getInt(i));
	 					}
	 				else
	 					break;
	 			}
	 		}
	 		else results.append("None ");
		 		 if (MesquiteDouble.isCombinable(Vh)) {
		 		results.append("\n\nCI/PI file columns\n\n");
	           	results.append("1.  Tip Name\n");
               	results.append("2.  X Value\n");
               	results.append("3.  Observed Y Value\n");
            	results.append("4.  Predicted Y Value (Yhat)\n");
            	results.append("5.  Lower " + width1*100 +"% Confidence Interval\n");
            	results.append("6.  Upper " + width1*100 + "% Confidence Interval\n");
            	results.append("7.  Lower " + width1*100 + "% Prediction Interval\n");
            	results.append("8.  Upper " + width1*100 + "% Prediction Interval\n");
            	results.append("9.  Lower " + width2*100 + "% Confidence Interval\n");
            	results.append("10.  Upper " + width2*100 + "% Confidence Interval\n");
            	results.append("11.  Lower " + width2*100 + "% Prediction Interval\n");
            	results.append("12.  Upper " + width2*100 + "% Prediction Interval\n\n\n");
		       	CIPIvalues(tree,obs1,obs2,tree.getRoot(),results,"\n"," ",true);
		    }
		    else results.append("No CI/PI values calculated - \n One or more undefined branch lengths\n\n");
		    return results.toString();
		}
	}   
    
    /**
     * Converts space characters to underscores '_'
     * @param nName String to convert, presumibly the name of a node
     * @return copy of nName with spaces converted
     */
    protected String spaceConvertString(String nName){
        int spacePos = nName.indexOf(" ");
        if (spacePos == -1)
            return nName;
        else {
            StringBuffer workBuffer = new StringBuffer(nName);
            while (spacePos != -1){
                workBuffer.setCharAt(spacePos,'_');
                spacePos = nName.indexOf(" ",spacePos+1);
            }
            return workBuffer.toString();
        }
    }
    

	
	/**
	@param lineEnding the end of line character sequence
	@param tableDelimiter the end of (table) column sequence
	@return String contains the CI table
	Entry point for generating the CI data as a string to write somewhere
	*/
	public String CIString(String lineEnding, String tableDelimiter, boolean convertSpaces) {
	    StringBuffer sb = new StringBuffer(1000);
	    CIPIvalues(tree,obs1,obs2,tree.getRoot(),sb,lineEnding,tableDelimiter,convertSpaces);
	    return sb.toString();
	}
	
	/*-----------------------------------------*/
	/** Returns results as brief String for use in legends **/
	public String getLegendText(){
		StringBuffer results = new StringBuffer(200);
		
	    results.append("Slope (least squares):            " + ls + "\n");
	    results.append("Standard Error:                  " + slopeSE + "\n");
	    results.append("Rates of Evolution:                      " + variance[n1] + " " + variance[n2] + "\n");
	    results.append("Estimated Root Node Values:              " + mean[n1] + "  " + mean[n2] + "\n");
	    results.append("Standard Errors of Root Node Estimates:  " + se[n1] + " " + se[n2] + "\n");
		return results.toString();
	}

	/** This just dispatches on drawingOption to draw either the cross hairs or the CI and PI lines. */
	public void drawOnChart(Graphics g, MesquiteChart chart){
		switch (drawingOption) {
			case ROOTCI: { 
				drawRootCiOnChart(g,chart);
				break;}
			case REGRESSIONCIPI: {
				drawRegressionCiPiOnChart(g,chart);
				break;}
		}
	}
			
	/**	 This plots the reconstructed root values and associated confidence intervals on top of the tips plot.  */
	private void drawRootCiOnChart(Graphics g, MesquiteChart chart) {
		Charter c = chart.getCharter();
		final int BRACKETWIDTH = 5;   //half width of the brackets at the end of error bars

		if (c == null)
			return;		
		// cache the root's position on the graph
		int rootX = c.xToPixel(mean[0],chart);
		int rootY = c.yToPixel(mean[1],chart);
		
		// start by drawing error bars for the CI  in both dimensions
		int xLowerBound = c.xToPixel(mean[0]-ci[0],chart);
		int xUpperBound = c.xToPixel(mean[0]+ci[0],chart);
		int yLowerBound = c.yToPixel(mean[1]-ci[1],chart);
		int yUpperBound = c.yToPixel(mean[1]+ci[1],chart);
		
		g.setColor(Color.green);
		g.drawLine(xLowerBound,rootY,xUpperBound,rootY);
		g.drawLine(xLowerBound,rootY-BRACKETWIDTH,xLowerBound,rootY+BRACKETWIDTH);
		g.drawLine(xUpperBound,rootY-BRACKETWIDTH,xUpperBound,rootY+BRACKETWIDTH);
		g.drawLine(rootX,yLowerBound,rootX,yUpperBound);
		g.drawLine(rootX-BRACKETWIDTH,yLowerBound,rootX+BRACKETWIDTH,yLowerBound);
		g.drawLine(rootX-BRACKETWIDTH,yUpperBound,rootX+BRACKETWIDTH,yUpperBound);
		
		// finally put up a small circle or rectangle in the root's location itself
		g.setColor(Color.red);
		// not yet sure how to do this, so I'll just draw a small rectangle
		g.drawLine(rootX-1,rootY-1,rootX+1,rootY-1);
		g.drawLine(rootX-1,rootY,rootX+1,rootY);
		g.drawLine(rootX-1,rootY+1,rootX+1,rootY+1);
	}
	
	/**  This plots the CI and PI lines */
	private void drawRegressionCiPiOnChart(Graphics g, MesquiteChart chart){
		Charter c = chart.getCharter();
		NumberArray Xpoints = chart.getXArray().cloneArray();
		if (c == null)
			return;
		if (!MesquiteDouble.isCombinable(Vh))
		   return;
		int startX = c.xToPixel(min[0],chart);     
		int startYOLS =  c.yToPixel(ls*min[0]+lsy,chart);
		int endX = c.xToPixel(max[0],chart);
		int endYOLS = c.yToPixel(ls*max[0]+lsy,chart);
		int extraLineWidth = 1;
		int oldX = MesquiteInteger.unassigned;
		int oldYLow = MesquiteInteger.unassigned;
		int oldYHigh = MesquiteInteger.unassigned;
		
		g.setColor(Color.red);
		oldX = MesquiteInteger.unassigned;
		Xpoints.sortUp();
		for (int i=0;i<chart.getNumPoints();i++) {
			// plot Confidence Interval line (segments) for width1
			double curX = Xpoints.getDouble(i);
			if (curX != MesquiteDouble.unassigned) {
				double yHat = ls*curX+lsy;
				double yCWidth = calcCI(curX,0,width1);
				//if the CI isn't calculable, set the width to zero and the main line will overwrite
				if (!MesquiteDouble.isCombinable(yCWidth))
					yCWidth = 0;
				int plotX = c.xToPixel(curX,chart);
				int plotHigh = c.yToPixel((yHat + yCWidth),chart);
				int plotLow = c.yToPixel((yHat - yCWidth),chart);
				showLineSegment(g,oldX,oldYHigh,plotX,plotHigh,oldYLow, plotLow);
				oldX = plotX;
				oldYLow = plotLow;
				oldYHigh = plotHigh;
			}		
		}
		g.setColor(Color.green);
		oldX = MesquiteInteger.unassigned;
		for (int i=0;i<chart.getNumPoints();i++) {
			// plot Prediction Interval line (segments) for width1
			double curX = Xpoints.getDouble(i);
			if (curX != MesquiteDouble.unassigned) {
				double yHat = ls*curX+lsy;
				double yPWidth = calcPI(curX,0,width1);
				//if the PI isn't calculable, set the width to zero and the main line will overwrite
				if (!MesquiteDouble.isCombinable(yPWidth))
					yPWidth = 0;
				int plotX = c.xToPixel(curX,chart);
				int plotHigh = c.yToPixel((yHat + yPWidth),chart);
				int plotLow = c.yToPixel((yHat - yPWidth),chart);
				showLineSegment(g,oldX,oldYHigh,plotX,plotHigh,oldYLow, plotLow);
				oldX = plotX;
				oldYLow = plotLow;
				oldYHigh = plotHigh;
			}		
		}
		
		g.setColor(Color.orange);
		oldX = MesquiteInteger.unassigned;
		Xpoints.sortUp();
		for (int i=0;i<chart.getNumPoints();i++) {
			// plot Confidence Interval line (segments) for width2
			double curX = Xpoints.getDouble(i);
			if (curX != MesquiteDouble.unassigned) {
				double yHat = ls*curX+lsy;
				double yCWidth = calcCI(curX,0,width2);
				//if the CI isn't calculable, set the width to zero and the main line will overwrite
				if (!MesquiteDouble.isCombinable(yCWidth))
					yCWidth = 0;
				int plotX = c.xToPixel(curX,chart);
				int plotHigh = c.yToPixel((yHat+yCWidth),chart);
				int plotLow = c.yToPixel((yHat-yCWidth),chart);
				showLineSegment(g,oldX,oldYHigh,plotX,plotHigh,oldYLow, plotLow);
				oldX = plotX;
				oldYLow = plotLow;
				oldYHigh = plotHigh;
			}		
		}
		g.setColor(Color.cyan);
		oldX = MesquiteInteger.unassigned;
		for (int i=0;i<chart.getNumPoints();i++) {
			// plot Prediction Interval line (segments) for width2
			double curX = Xpoints.getDouble(i);
			if (curX != MesquiteDouble.unassigned) {
				double yHat = ls*curX+lsy;
				double yPWidth = calcPI(curX,0,width2);
				//if the PI isn't calculable, set the width to zero and the main line will overwrite
				if (!MesquiteDouble.isCombinable(yPWidth))
					yPWidth = 0;
				int plotX = c.xToPixel(curX,chart);
				int plotHigh = c.yToPixel((yHat + yPWidth),chart);
				int plotLow = c.yToPixel((yHat - yPWidth),chart);
				showLineSegment(g,oldX,oldYHigh,plotX,plotHigh,oldYLow, plotLow);
				oldX = plotX;
				oldYLow = plotLow;
				oldYHigh = plotHigh;
			}		
		}
		
		g.setColor(Color.black);
		// plot the main regression line and thicken it
		g.drawLine(startX, startYOLS,endX,endYOLS);
		if (Math.abs(endYOLS-startYOLS)>(endX-startX))   // abs(slope) > 1
			g.drawLine(startX+extraLineWidth,startYOLS,endX+extraLineWidth,endYOLS);
		else g.drawLine(startX,startYOLS+extraLineWidth,endX,endYOLS+extraLineWidth);
			
        //Plot the -10..10 points and any userdefined points			
        for(int i = -10; i <= -1; i++){
            double x = min[0] + min[0]*i/100.0;
 	        double yHat = ls*x + lsy;
 	        showPoint(x,yHat,chart,g);
	    } 
	    for(int i = 1; i <= 10; i++){
            double x = max[0]+max[0]*i/100.0;
 	        double yHat = ls*x + lsy;
 	        showPoint(x,yHat,chart,g);
	    } 
        for (int i=0; i < CIPIValues.size(); i++){
	        double x = CIPIValueAt(i).getValue();
		 	double yHat = ls*x + lsy;
	        showPoint(x,yHat,chart,g);
        }
	}  /* end of drawRegressionCiPiOnChart() */
	
	/** Result of refactoring */
	private void showLineSegment(Graphics g, int oldX, int oldYHigh, int plotX, int plotHigh, int oldYLow, int plotLow) {
		if (MesquiteInteger.isCombinable(oldX)) {
			g.drawLine(oldX,oldYHigh,plotX,plotHigh);
			g.drawLine(oldX,oldYLow,plotX,plotLow);
		}
	}
	
	/** Draw CI/PI line segments (despite method name) for values that don't correspond to tips */
	private void showPoint(double x, double yHat,MesquiteChart chart, Graphics g) {
		Charter c = chart.getCharter();
		double yCWidth = calcCI(x,0,width1);
		int plotX = c.xToPixel(x,chart);
		int plotLeft = plotX-1;
		int plotYHat = c.yToPixel(yHat,chart);
 		double highY = yHat + yCWidth;
		double lowY = yHat - yCWidth;
		int plotHigh = c.yToPixel(highY,chart);
		int plotLow = c.yToPixel(lowY,chart);
        g.setColor(Color.darkGray);
 	    g.drawLine(plotLeft,plotYHat,plotX,plotYHat);
 		g.setColor(Color.red);
		g.drawLine(plotLeft,plotHigh,plotX,plotHigh);
		g.drawLine(plotLeft,plotLow,plotX,plotLow);
        double yPWidth = calcPI(x,0,width1);
 		highY = yHat + yPWidth;
		lowY = yHat - yPWidth;
		plotHigh = c.yToPixel(highY,chart);
		plotLow = c.yToPixel(lowY,chart);
		g.setColor(Color.green);
		g.drawLine(plotLeft,plotHigh,plotX,plotHigh);
		g.drawLine(plotLeft,plotLow,plotX,plotLow);
		yCWidth = calcCI(x,0,width2);
 		highY = yHat + yCWidth;
		lowY = yHat - yCWidth;
		plotHigh = c.yToPixel(highY,chart);
		plotLow = c.yToPixel(lowY,chart);
		g.setColor(Color.orange);
		g.drawLine(plotLeft,plotHigh,plotX,plotHigh);
		g.drawLine(plotLeft,plotLow,plotX,plotLow);
		yPWidth = calcPI(x,0,width2);
 		highY = yHat + yPWidth;
		lowY = yHat - yPWidth;
		plotHigh = c.yToPixel(highY,chart);
		plotLow = c.yToPixel(lowY,chart);
		g.setColor(Color.cyan);
 	    g.drawLine(plotLeft,plotHigh,plotX,plotHigh);
		g.drawLine(plotLeft,plotLow,plotX,plotLow);	
	}


	public boolean doCalculations(NumberArray xArray, NumberArray yArray, boolean screen15, MesquiteChart chart, CommandRecord commandRec) {
		// TODO Auto-generated method stub
		return false;
	}
}

