
package mesquite.ac3.lib;

import mesquite.categ.lib.*;
import mesquite.cont.lib.*;
import mesquite.lib.MesquiteDouble;
import mesquite.lib.MesquiteInteger;
import mesquite.lib.MesquiteMessage;
import mesquite.lib.MesquiteTree;
import mesquite.lib.Tree;
import mesquite.lib.characters.CharacterDistribution;

public abstract class AC32CTStatPak extends AC3StatPak {

	public static final int ROOTCI	= 0;

	protected NRtDistribution       	tStats;             //an object for calculating t statistics from jsci[Numerical Recipes?]
	protected NoPolyTree		  		tree;			 	//putting the tree here and handling the contrast calculations
	 									 				//internally seems to be the only way to implement 
	//ajout
	protected CategoricalDistribution    obs1;	           //more caching
	
	protected ContinuousDistribution    obs2;
	protected CharacterDistribution		obs3;
	protected MesquiteInteger           dfReduce;       //User specified amount to reduce degrees of freedom
    protected int         			    root;                       //so we aren't always asking the tree...
	protected double[]   				ci;  	  	                 //Confidence intervals around root reconstruction
	protected double[]			se;
	protected double  			sumresid;   	                 //sum of regression residuals
	protected double     		slopeCI;    	                 //Confidence interval around least squares slope
	protected double     		interCI;    	                 //Confidence interval around least squares intercept
	protected double 			slopeSE;    	                 //Standard error of least squares slope
	protected double     		interSE;		                 //Standard error of least squares intercept
	protected double   		      v1;                           // branch length adjustment factor 
	protected double     		v2;                           // branch length adjustment factor
	protected double  			v_adjust;                  // == (v1+v2)/(v1*v2)
	protected double     		x2Sum;
	protected double 			sigma2y;
	protected double     		Vh;				//height of added tips for PI calculations 
	protected int				polytomies; 	                  //number of polytomies (ZLB's) in the tree
	protected boolean		  	contempTips;                   //equals True if tips are contemporaneous
	protected int 	         	drawingOption = ROOTCI;  // might be overwritten to REGRESSIONCIPI

	
	
	public AC32CTStatPak(MesquiteInteger dfr) {
		super();
		ci = new double[] {0.0, 0.0}; 
		se = new double[] {0.0, 0.0};
        dfReduce = dfr;
	}
	
	
	
	/*-----------------------------------------*/
	/** Set the tree so we can do the contrast calculations internally */
	public void setTree(Tree tree) {
		if (tree instanceof NoPolyTree)
			this.tree = (NoPolyTree)tree;
		else 
			this.tree = new NoPolyTree((MesquiteTree)tree);
		root = tree.getRoot();
		clearCIPIlist();
	}

	
	
	/*-----------------------------------------*/
	/** Set the dataset for the first variable */
	public void setObserved1(CategoricalDistribution observed) {
		obs1 = observed;
		clearCIPIlist();
	}
	
	/*-----------------------------------------*/
	/** Set the dataset for the second variable */
	public void setObserved2(ContinuousDistribution observed) {
		obs2 = observed;
		clearCIPIlist();
	}
	
	/*-----------------------------------------*/
	/** Set the dataset for the independant variable */
	public void setObserved3(CharacterDistribution observed) {
		obs3 = observed;
		clearCIPIlist();
	}

	/*-----------------------------------------*/
  	public abstract void querySetVh();

	/*-----------------------------------------*/
  	/** Vh is used for calculating a prediction interval for an added tip taxon.  It
  	 is the length of the branch connecting the added tip (directly) to the root. */
  	public double getVh(){
  		return Vh;
  	}

  	/*-----------------------------------------*/
  	/** Vh is used for calculating a prediction interval for an added tip taxon.  It
  	 is the length of the branch connecting the added tip (directly) to the root. */
  	public void setVh(double tmp){
  		if (MesquiteDouble.isCombinable(tmp))
  			Vh = tmp;	
  	}	

	public abstract void setVhToMeanHeight(Tree tree);


	/*-----------------------------------------*/
	/** This returns the user specified number of degrees of freedom reduction.  */	
	public int getDFReduction() {
		return dfReduce.getValue(); 
	}
    
	/*-----------------------------------------*/
	/** This sets both the user specified reduction in degrees of freedom */
	public void setDFReduction(int r){
		if (MesquiteInteger.isCombinable(r)) {
			dfReduce.setValue(r);
		}
	}
    
    public int getDegreesOfFreedom(){
        return df;
    }
    /*-----------------------------------------*/
    /**  After getting the number of zero-length branches, this method prompts the user
     to estimate the number of degrees of freedom lost to the polytomies represented by those
     zero length branches.  Need to say something about Diaz-Uriarte and Garland 1999. */
    public void queryDFReduce(Tree tree, boolean notifyIfZero){
        countInternalNodes(tree,tree.getRoot());  // count the internal nodes, to set number
        int polytomies = countZLBs(tree,tree.getRoot());     // Zero-Length-Branches are what AC3 calls "polytomies", count to inform the user
        if ((polytomies >0)) {
            int tmp = MesquiteInteger.queryInteger(null,  
                                                   "DF Reduce ",
                                                   "This tree contains polytomies, indicated by the presence \n" +
                                                   "of " + polytomies + " zero-length branch(es).\n" +
                                                   "If some of these collapsed branches indicate\n" +
                                                   "soft polytomies, then you may want to reduce\n" +
                                                   "degrees of freedom (see Purvis and Garland, 1993,\n" +
                                                   "Syst. Biol. 42:569-575; Garland and Diaz-Uriarte, 1999,\n" +
                                                   "Syst. Biol. 48:547-558).\n\n " +
                                                   "Enter number of Degrees of Freedom to subtract for soft polytomies",
                                                   getDFReduction(),
                                                   0,
                                                   MesquiteInteger.infinite);
                                                    
            if (MesquiteInteger.isCombinable(tmp)) {
                setDFReduction(tmp);
            }
        }
        else {
            if (notifyIfZero) {
                int tmp = MesquiteInteger.queryInteger(null,  
                                                       "DF Reduce ",
                                                       "This tree is fully dichotomized and contains no polytomies.\n" +
                                                       "If, for whatever reason, you wish to decrease\n" +
                                                       "degrees of freedom (see Purvis and Garland, 1993,\n" +
                                                       "Syst. Biol. 42:569-575; Garland and Diaz-Uriarte, 1999,\n" +
                                                       "Syst. Biol. 48:547-558), you may do so here.\n\n" +
                                                       "Enter degrees of freedom to subtract",
                                                       getDFReduction(),
                                                       0,
                                                       MesquiteInteger.infinite);
                //alert(null,"","The tree has no polytomies, hence there is no need to reduce degrees of freedom to compensate for polytomies.");
                if (MesquiteInteger.isCombinable(tmp)) 
                    setDFReduction(tmp);
                 else
                    setDFReduction(0);
            }
        }
    }

    /*-----------------------------------------*/
    /** This method counts the zero-length-branches in the tree.  Because this is AC3,
    it needs to look for zero length branches, rather than polytomies.  To make sure
    this works, the method checks that it is getting an instance of NoPolyTree. */
    public int countZLBs(Tree tree, int node) {
        if (!(tree instanceof NoPolyTree)) {
            MesquiteMessage.warnProgrammer("Tree is not a NoPolyTree - counting 0 polytomies");
            return 0;
        }
        polytomies = countZLBsHelper(tree,node); 
        return polytomies;
    }

    /* This is a helper function */
    private int countZLBsHelper(Tree tree, int node) {
        if(tree.nodeIsInternal(node)) {
            int count = 0;
            if ((tree.getBranchLength(node) == 0.0) && (node != tree.getRoot())) { // AC3 roots are ZLB's (arbitrarily)
                    count++;
                }
            for(int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter);daughter=tree.nextSisterOfNode(daughter))
                count += countZLBsHelper(tree, daughter);
            return count;
        }
        else
            return 0;
    }
    
    public String CIString(String lineEnding, String tableDelimiter,boolean convertSpaces){
        return "";
    }
    


}

