/* Mesquite ~~ Copyright 1997-2003 W. & D. Maddison*/
/* This java translation and commentary copyright 2001-2003 Peter E. Midford */

package mesquite.ac3.lib;

import mesquite.lib.*;

// Statistics (t and someday F) code below this line.  Much of this code is 
// taken from Press, W. H., Flannery, B. P., Teukolsky, S. A., and         
// Vetterling, W. T. 1989. Numerical Recipes in Pascal. Cambridge, Cambridge
// University Press.  The results produced by these routines differ        
// from the values produced by SPSS/PC+ Version 5.0.  The values usually    
// differ in the fourth decimal place.  We also compared results from the   
// the routines used herein (from Numerical Recipes) with those produced by 
// an HP-21S calculator; they agree to more than 7 decimal places.          
 
// Since these methods use code from Numerical Recipes, I'll leave them here,
// buried in the AC3 package, which should reduce the chance of theme being
// redistributed in a manner that violates someone's copyright.  At some point,
// we may want to replace these, since they seem to disagree with other sources,
// such as JSCI.  They are here now specifically for compatibility with the 
// original AC3 program.  Updating this may have to wait for an update to the
// DOS version of AC3.

// All code not taken from Numerical Recipes in Pascal is copyright 1996
// by Peter E. Midford, assigned to Theodore A. Garland Jr.
// Java translation copyright 2001 Peter E. Midford


public class NRtDistribution {
	private int df;
	
	public NRtDistribution(int DoF) {
		if(DoF<=0)
			MesquiteMessage.printStackTrace("Error - nonpositive degrees of freedom for t-distribution.");
		df = DoF;
	}
	
	public int getdf() {
		return df;
	}
	
	

    // This method computes the natural log of the gamma function.  It is
    // translated from Pascal code on page 177 of Numerical Recipes in Pascal. 
    // The authors claim the results from this function are good to 2e-10. 

	private double gammln(double xx) {
    	final double stp = 2.50662827465;
		double x;
		double tmp;
		double ser;

        x = xx - 1.0;
        tmp = x + 5.5;
        tmp = (x+0.5)*Math.log(tmp)-tmp;
        ser = 1.0+76.18009173/(x+1.0)-86.50532033/(x+2.0)+
                   24.01409822/(x+3.0)-1.231739516/(x+4.0)+
                   0.120858003e-2/(x+5.0)-0.536382e-5/(x+6.0);
        return tmp+Math.log(stp*ser);
    }
    
    // This method computes the continued fraction helper function for 
    // the incomplete beta function.  It comes from pages 188-189 of        
    // Numerical Recipes in Pascal.                                       
    private double betacf(double a, double b, double x) {
		int itmax = 100;
        double eps = 3.0e-7;
   		double tem;
   		double qap;
   		double qam;
   		double qab;
   		double em;
   		double d;
        double bz;
        double bpp;
        double bp;
        double bm = 1.0;
        double az = 1.0;
        double app;
        double am = 1.0;
        double aold = 1.0;  // keep java initialization check happy
        double ap;
        
        qab = a + b;
        qap = a + 1;
        qam = a - 1;
        bz = 1.0-qab*x/qap;
        for (int m = 1; m <= itmax; m++) { 
        	em = m;
            tem = em+em;
            d = em*(b-m)*x/((qam+tem)*(a+tem));
            ap = az+d*am;
            bp = bz+d*bm;
            d = -(a+em)*(qab+em)*x/((a+tem)*(qap+tem));
            app = ap+d*az;
            bpp = bp+d*bz;
            aold = az;
            am = ap/bpp;
            bm = bp/bpp;
            az = app/bpp;
            bz = 1.0;
            if (Math.abs(az-aold) < eps*Math.abs(az)) break;
      	}
        if (Math.abs(az-aold) >= eps*Math.abs(az)) {
        	MesquiteMessage.warnProgrammer("Error in RootStat.betacf");
            MesquiteMessage.printStackTrace("a or b too big, or itmax too small");
        }
        return az;
     }
    
	// This method calculates the incomplete beta function.  It is translated from  
	// Pascal code on page 188 of Numerical Recipes in Pascal.
	private double betai(double a, double b, double x) {
		double bt;
		
        if ((x < 0.0) || (x > 1.0)) 
            Debugg.printStackTrace("X (" + x +") out of range [0-1] in RootStat.betai");
        if ((x == 0.0) || (x == 1.0)) 
        	bt = 0.0;
        else
            bt = Math.exp(gammln(a+b)-gammln(a)-gammln(b)
                   +a*Math.log(x)+b*Math.log(1-x));
        if (x < (a+1)/(a+b+2))
            return bt*betacf(a,b,x)/a;
        else
            return 1.0-bt*betacf(b,a,1-x)/b;
      } 

	// This method calculates the p-value for the t-statistic t with df      
	// degrees of freedom.                                             
    public double cumulative(double t) {
        if (t < 0)
            return betai(df/2.0,0.5,df*1.0/(df*1.0+t*t))/2.0;
        else
            return 1-betai(df/2.0,0.5,df*1.0/(df*1.0+t*t))/2.0; 
    }

	// This function finds the roots of fx in the interval [x1,x2] within 
    // the tolerance tol.  It is translated from the zbrent algorithm listed 
    // on pp 285-256 of Numerical Recipes in Pascal, except where I cheated
    // and hardcoded calls to tvalue in the appropriate places in the algorithm, 
    // while the original passed in a function.  Since Java doesn't have functional
    // arguments, we would need to do something else to extended it for other CI's
    private double zbrent(double x1, double x2, double tol, double ci) {
		int itmax = 100;
        double eps = 3.0e-8;
        double a = x1;
        double b = x2;
        double c = x1;    //these assignments are only to satisfy the java compiler
        double d = b-a;   //however, the values are correct as fb*fb is >0, or unless
        double e = d;     // b *is* the root, when these values shouldn't matter
        double min1;
        double min2;
        double min;
        double fa;
        double fb;
        double fc;
        double p;
        double q;
        double r;
        double s;
        double tol1;
        double xm;

        a = x1;
        b = x2;
        fa = cumulative(a)-ci; // hardcoded method name
        fb = cumulative(b)-ci; // hardcoded methdo name
        if (fb*fa > 0.0) {
        	MesquiteMessage.warnProgrammer("Error in routine RootStat.zbrent");
            MesquiteMessage.warnProgrammer("root must be bracketed");
        }
        fc = fb;
        for (int iter= 1; iter <= itmax; iter++) {
        	if (fb*fc > 0.0) {
            	c = a;
                fc = fa;
                d = b-a;
                e = d;
            }
            if (Math.abs(fc) < Math.abs(fb)) {
            	a = b;    // swap everything.  The following assignment works because the
                b = c;    // previous condition is always true the first time.
                c = a;
                fa = fb;
                fb = fc;
                fc = fa;
			}
            tol1 = 2.0*eps*Math.abs(b)+0.5*tol; // Convergence check 
            xm = 0.5*(c-b);
            if ((Math.abs(xm) <= tol1) && (Math.abs(fa) > Math.abs(fb)))
            	return b;
            if ((Math.abs(e) >= tol1) && (Math.abs(fa) > Math.abs(fb))) {
            	s = fb/fa;   // Attempt inverse quadratic interpolation
                if (a == c) {
                	p = 2.0*xm*s;
                    q = 1.0 - s;
                }
                else {
                	q = fa/fc;
                    r = fb/fc;
                    p = s*(2.0*xm*q*(q-r)-(b-a)*(r-1.0));
                    q = (q-1.0)*(r-1.0)*(s-1.0);
                }
                if (p > 0.0)
                	q = -q;
                p = Math.abs(p);
                min1 = 3.0*xm*q-Math.abs(tol1*q);
                min2 = Math.abs(e*q);
                if (min1 < min2) min = min1; 
                else min = min2;
                if (2.0*p < min) {
                	e = d;         // Accept interpolation
                    d = p/q;
                }
                else {
                	d = xm;        // Interpolation failed, use bisection
                    e = d;
                }
            }
            else  {
            	d = xm;      // Bounds decreasing too slowly, use bisection
                e = d;
            }
            a = b;          // Move last best guess to a
            fa = fb;
            if (Math.abs(d) > tol1) 
            	b = b+d;
            else {
            	if (xm >= 0)  b = b + Math.abs(tol1);
                else b = b-Math.abs(tol1);
            }
            fb = cumulative(b)-ci;     // another hardcoded function
        }
        MesquiteMessage.warnProgrammer("Error in routine RootStat.zbrent");
        MesquiteMessage.warnProgrammer("maximum number of iterations exceeded");
        return  b;
    }



	// This function calculates confidence intervals for df degrees of      
	// freedom with the width CI.  The approach I took was to solve for an x 
	// whose t value equaled the desired CI width.  To solve the equation I 
	// used the zbrent algorithm listed on pp 285-256 of Numerical Recipes in
	// Pascal.                                          

	public double CI(double CI) {
        return zbrent(0,100,0.00000001,CI);
	}
	
	}
 

