/* Mesquite ~~ Copyright 1997-2006 W. & D. Maddison*/
/* This java translation and commentary copyright 2001-2006 P. E. Midford */
/* This file copyright 2003-2006 P. E. Midford */

package mesquite.ac3.lib;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import mesquite.lib.*;
import mesquite.cont.lib.*;


public class UnivariateContrastCalculator extends ContrastCalculator{	

	/** Constructor  for a simple calculator 
	 * @param nodesToRecalc */

	public UnivariateContrastCalculator(Tree localTree, ContinuousDistribution observedStates) {
		super();
		if (localTree == null || observedStates == null) {
			MesquiteMessage.printStackTrace("A null value was passed to the constructor for UnivariateContrastCalculator");
			tree = null;
			observedStates = null;
		}
		else {
			tree = localTree;
			states = observedStates;
			numNodes = tree.getNumNodeSpaces();
			checkArrays();
		}	
	}


	/** this tries to reuse as much memory as possible, while resetting everything */
	public void reinitialize(Tree theTree, boolean[] theDeletions,ContinuousDistribution observedStates) {
		tree = theTree; //reinitialize internals
		states = observedStates;
		numNodes = tree.getNumNodeSpaces(); 
		checkArrays(); // either blank all arrays or allocate new ones if space requirements change
	}

	/*..................................................................................................................*/
	/** fills or updates the internal arrays */
	protected void checkArrays() {
		if (contrasts == null || contrasts.length != numNodes) {
			contrasts = new double[numNodes];
			rawContrasts = new double[numNodes];
			vPrime = new double[numNodes];
			x = new double[numNodes];
			stdDev = new double[numNodes];
			height = new double[numNodes];
			weightPassThrough = new double[numNodes];
		}
		else {
			DoubleArray.deassignArray(contrasts);
			DoubleArray.deassignArray(rawContrasts);  //should this be zeroArray?
			DoubleArray.zeroArray(x);
			DoubleArray.zeroArray(vPrime);
			DoubleArray.zeroArray(stdDev);
			DoubleArray.deassignArray(weightPassThrough);
		}
	}

	/*.................................................................................................................*/
	// Added checks for missing data and handle the cases of interior nodes with one or both daughters
	// missing data.
	protected void vCalculation(int node) {
		for (int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter))
			vCalculation(daughter);
		if (tree.nodeIsInternal(node)) {
			int left = tree.firstDaughterOfNode(node);
			int right = tree.lastDaughterOfNode(node);
			if (MesquiteDouble.isCombinable(vPrime[left]) && MesquiteDouble.isCombinable(vPrime[right]))  {// both good
				vPrime[node] = branchWeight(node) + (vPrime[left] * vPrime[right]) / (vPrime[left] + vPrime[right]); /*vPrime used temporarily!*/
				stdDev[node] = Math.sqrt(vPrime[right] + vPrime[left]);				
			}
			else if (MesquiteDouble.isCombinable(vPrime[left])) { // only left good
				weightPassThrough[node] = branchWeight(left);
				//vPrime[left] += branchWeight(node);
				vPrime[node] = branchWeight(node)+vPrime[left];
				stdDev[node] = MesquiteDouble.unassigned;
			}
			else if (MesquiteDouble.isCombinable(vPrime[right])){ // only right good
				weightPassThrough[node] = branchWeight(right);
				//vPrime[right] += branchWeight(node);
				vPrime[node] = branchWeight(node) + vPrime[right];
				stdDev[node] = MesquiteDouble.unassigned;
			}
			else { // both bad  - so the whole subtree below this node is dead?
				vPrime[node] = MesquiteDouble.unassigned;
				stdDev[node] = MesquiteDouble.unassigned;
			}
		}		
		else { // Check that the node actually has data available
			if (states != null  && 
					MesquiteDouble.isCombinable(x[node]=states.getState(tree.taxonNumberOfNode(node), 0))) {  
				//x[node] = states.getState(tree.taxonNumberOfNode(node), 0);
				vPrime[node] = branchWeight(node);
			}
			else{  //bad data
				x[node] = MesquiteDouble.unassigned;
				vPrime[node] = MesquiteDouble.unassigned;
			}
		}
	}
	/*......................................................................................................................*/
	protected void heightCalculation(double ht, int node, int mode) {
		double myHeight;
		if (tree.nodeIsInternal(node)) {
			if (MesquiteDouble.isCombinable(stdDev[node])) { // good data
				if (node != tree.getRoot())
					if (mode == ContrastForChar.RAWHEIGHT)
						myHeight = ht + branchWeight(node);
					else
						myHeight = ht + vPrime[node];
				else    
					myHeight = 0;
				height[node] = myHeight;
			}
			else {
				height[node] = MesquiteDouble.unassigned;
				if (mode == ContrastForChar.RAWHEIGHT)
					myHeight = ht + branchWeight(node);
				else
					myHeight = ht + branchWeight(node);  //vPrime[node];
			}
			for (int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter))
				heightCalculation(myHeight, daughter,mode);
		}
		else {  // a tip
			if (MesquiteDouble.isCombinable(x[node])) {
				if (mode == ContrastForChar.RAWHEIGHT)
					height[node] = ht + branchWeight(node);
				else
					height[node] = ht + vPrime[node];
			}
			else 
				height[node] = MesquiteDouble.unassigned;		
		}
	}

	/*.................................................................................................................*/
	protected void xCalculation(int node) {
		for (int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter))
			xCalculation(daughter);
		if (tree.nodeIsInternal(node)) {
			int left = tree.firstDaughterOfNode(node);
			int right = tree.lastDaughterOfNode(node);
			if (MesquiteDouble.isCombinable(x[left]) && MesquiteDouble.isCombinable(x[right])) {  // good contrast
				x[node] = (vPrime[right]*x[left] + vPrime[left]*x[right] )/(vPrime[left] +vPrime[right]);  
				rawContrasts[node] =(x[left] - x[right]);
				contrasts[node] = rawContrasts[node]/stdDev[node];
			}
			else {
				rawContrasts[node] = MesquiteDouble.unassigned;
				contrasts[node] = MesquiteDouble.unassigned;
				x[node] = MesquiteDouble.unassigned;  // Default unless either one of left or right is good
				if(MesquiteDouble.isCombinable(x[right]))
					x[node] = x[right];  // no value from left to average with, so all we can do is propagate down!
				if(MesquiteDouble.isCombinable(x[left]))
					x[node] = x[left];  // no value from right to average with, so all we can do is propagate down!
			}
		}
		else {
			if (states != null  && MesquiteDouble.isCombinable(states.getState(tree.taxonNumberOfNode(node), 0)))  
				x[node] = states.getState(tree.taxonNumberOfNode(node), 0); 
			else
				x[node] = MesquiteDouble.unassigned;
			stdDev[node] = MesquiteDouble.unassigned;
			contrasts[node] = MesquiteDouble.unassigned;
		}
	}


	/** This also assumes things have been set up. */
	public double [] heightCalculation(int mode){
		int root = tree.getRoot();
		vCalculation(root);
		heightCalculation(0.0,root,mode);
		return height;
	}

	/** This copies tip values into the x array 
	@return double [] - the array of tip values, internal node values are set to MesquiteDouble.unassigned */
	public double [] fillTips(int node) {
		for (int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter))
			fillTips(daughter);
		if (tree.nodeIsInternal(node)) {
			x[node] = MesquiteDouble.unassigned;

		}
		else {
			x[node] = states.getState(tree.taxonNumberOfNode(node), 0);  
		}
		return x;
	}
public double [] DevExtendedCalculation(NumberArray nodesToRecalc){
	double[] finalContrasts=null;
	if(MesquiteDouble.isCombinable(nodesToRecalc.getDouble(0))){
		List modifiedContrasts=new ArrayList();
		for(int i=0;i<nodesToRecalc.getSize();i+=3){
			boolean leftNodeIsTerm =tree.nodeIsTerminal(nodesToRecalc.getInt(i+1));	
			boolean rightNodeIsTerm =tree.nodeIsTerminal(nodesToRecalc.getInt(i+2));	
			if(nodesToRecalc.getInt(i)!=0){
				double length=0;
						if(leftNodeIsTerm && rightNodeIsTerm){ //contrast between two terminal taxa
					length=getHeightToRoot(tree,nodesToRecalc.getInt(i+1))+getHeightToRoot(tree,nodesToRecalc.getInt(i+2))-2*getHeightToRoot(tree,nodesToRecalc.getInt(i));
				}
				else if(!leftNodeIsTerm && rightNodeIsTerm){//contrast between an internal node and a terminal taxa
				length=vPrime[nodesToRecalc.getInt(i+1)]+getHeightToRoot(tree,tree.motherOfNode(nodesToRecalc.getInt(i+1)))+getHeightToRoot(tree,nodesToRecalc.getInt(i+2))-2*getHeightToRoot(tree,nodesToRecalc.getInt(i));
				}
				else if(leftNodeIsTerm && !rightNodeIsTerm){//contrast between an internal node and a terminal taxa
					length=vPrime[nodesToRecalc.getInt(i+2)]+getHeightToRoot(tree,tree.motherOfNode(nodesToRecalc.getInt(i+2)))+getHeightToRoot(tree,nodesToRecalc.getInt(i+1))-2*getHeightToRoot(tree,nodesToRecalc.getInt(i));
				}
				else{ // contrast between two internal nodes
					length=vPrime[nodesToRecalc.getInt(i+2)]+vPrime[nodesToRecalc.getInt(i+1)]+getHeightToRoot(tree,tree.motherOfNode(nodesToRecalc.getInt(i+2)))+getHeightToRoot(tree,tree.motherOfNode(nodesToRecalc.getInt(i+1)))-2*getHeightToRoot(tree,nodesToRecalc.getInt(i));
				}
				length=Math.sqrt(length);
				modifiedContrasts.add(new Double(nodesToRecalc.getDouble(i)));
					modifiedContrasts.add(new Double(length));
				}
			else break;
		}
		finalContrasts=new double[modifiedContrasts.size()];
		Iterator iter=modifiedContrasts.iterator();
		int j=0;
		while(iter.hasNext()){
			finalContrasts[j]=Double.parseDouble(iter.next().toString());
			j++;
		}
	}
	return finalContrasts;
	}

	protected double [] xExtendedCalculation(NumberArray nodesToRecalc){
		double[] finalContrasts=null;
		if(MesquiteDouble.isCombinable(nodesToRecalc.getDouble(0))){
			List modifiedContrasts=new ArrayList();
			for(int i=0;i<nodesToRecalc.getSize();i+=3){
				boolean leftNodeIsTerm =tree.nodeIsTerminal(nodesToRecalc.getInt(i+1));	
				boolean rightNodeIsTerm =tree.nodeIsTerminal(nodesToRecalc.getInt(i+2));	
				if(nodesToRecalc.getInt(i)!=0){
					double length=0;
					double rawContrast=0;
					if(leftNodeIsTerm && rightNodeIsTerm){ //contrast between two terminal taxa
						//double state1=states.getState(tree.taxonNumberOfNode(nodesToRecalc.getInt(i+1)));
						//double state2=states.getState(tree.taxonNumberOfNode(nodesToRecalc.getInt(i+2)));
						//rawContrast=state1-state2;
						rawContrast=x[nodesToRecalc.getInt(i+1)]-x[nodesToRecalc.getInt(i+2)];
						length=getHeightToRoot(tree,nodesToRecalc.getInt(i+1))+getHeightToRoot(tree,nodesToRecalc.getInt(i+2))-2*getHeightToRoot(tree,nodesToRecalc.getInt(i));

					}
					else if(!leftNodeIsTerm && rightNodeIsTerm){//contrast between an internal node and a terminal taxa
						rawContrast=x[nodesToRecalc.getInt(i+1)]-x[nodesToRecalc.getInt(i+2)];
						length=vPrime[nodesToRecalc.getInt(i+1)]+getHeightToRoot(tree,tree.motherOfNode(nodesToRecalc.getInt(i+1)))+getHeightToRoot(tree,nodesToRecalc.getInt(i+2))-2*getHeightToRoot(tree,nodesToRecalc.getInt(i));
					}
					else if(leftNodeIsTerm && !rightNodeIsTerm){//contrast between an internal node and a terminal taxa
						rawContrast=x[nodesToRecalc.getInt(i+1)]-x[nodesToRecalc.getInt(i+2)];
						length=vPrime[nodesToRecalc.getInt(i+2)]+getHeightToRoot(tree,tree.motherOfNode(nodesToRecalc.getInt(i+2)))+getHeightToRoot(tree,nodesToRecalc.getInt(i+1))-2*getHeightToRoot(tree,nodesToRecalc.getInt(i));
					}
					else{ // contrast between two internal nodes
						rawContrast=x[nodesToRecalc.getInt(i+1)]-x[nodesToRecalc.getInt(i+2)];
						length=vPrime[nodesToRecalc.getInt(i+2)]+vPrime[nodesToRecalc.getInt(i+1)]+getHeightToRoot(tree,tree.motherOfNode(nodesToRecalc.getInt(i+2)))+getHeightToRoot(tree,tree.motherOfNode(nodesToRecalc.getInt(i+1)))-2*getHeightToRoot(tree,nodesToRecalc.getInt(i));
						/*	length=Math.sqrt(length);
					modifiedContrasts.add(new Double(nodesToRecalc.getDouble(i)));
					if(length!=0)
						modifiedContrasts.add(new Double(rawContrast/length));
					else
						modifiedContrasts.add(new Double(MesquiteDouble.unassigned));*/
					}
					length=Math.sqrt(length);
					modifiedContrasts.add(new Double(nodesToRecalc.getDouble(i)));
					if(length!=0)
						modifiedContrasts.add(new Double(rawContrast/length));
					else
						modifiedContrasts.add(new Double(MesquiteDouble.unassigned));
				}
				else
					break;
			}
			finalContrasts=new double[modifiedContrasts.size()];
			Iterator iter=modifiedContrasts.iterator();
			int j=0;
			while(iter.hasNext()){
				finalContrasts[j]=Double.parseDouble(iter.next().toString());
				j++;
			}
		}
		return finalContrasts;
	}

	/** Returns total height from root to node*/
	private  double getHeightToRoot(Tree t,int node) { 
		if(node==t.getRoot() || !t.nodeExists(node)) return 0;
		else return t.getBranchLength(node,1)+getHeightToRoot(t,t.motherOfNode(node));
	}
}	