/* AC3 Mesquite module */

package mesquite.ac3.PagelTransform;
/*~~  */

import mesquite.lib.*;
import mesquite.lib.duties.*;

/** ======================================================================== */
public class PagelTransform extends BranchLengthsAlterer {
	int [] depth;
	double resultNum;
	/*.................................................................................................................*/
	public boolean startJob(String arguments, Object condition, boolean hiredByName) {
		return true;
  	 }
  	 

	/*.................................................................................................................*/
	/**
     *
     * @param tree
     * @param resultString
     * @param commandRec
     * @return true if calling code is responsible for notifying listeners of tree 
	 */
    
    public boolean transformTree(AdjustableTree tree, MesquiteString resultString){
        doTransform(tree, tree.getRoot());  // this is now overloaded
		return true;
	}
	
    
    /**
     * This procedure sets all the branch lengths to those dictated by
     * Pagel's (1992, J. of Theoretical Biol 156:431-442) arbitrary method.
     * The depth of a node is equal to the maximum of the number of
     * bifurcation levels of the left child and the right child plus one.
     * If a node is a polytomy, its depth is set to the maximum of the
     * bifurcation levels of the offspring, but one is not added.  This
     * results in polytomies being assigned the least depth possible.  This
     * was a (AC3) internal decision, as Pagel(1992, p. 441 paragraph 7) does
     * not discuss how to assign a depth to a polytomy node.  
     * 
     * @param tree
     * @param root
     */
    private void doTransform(AdjustableTree tree, int root) {
        depth = new int[tree.getNumNodeSpaces()];
        pagelMe1(tree,root);
        pagelMe2(tree,root);
    }
        
    
    /**
     * 
     * @param tree
     * @param nd
     * @return
     */
    // This method computes the depths recursively.  The original
    // Pagelme procedure was split into 2 pieces to support AC3's implementation of polytomy.
    // PEM 9-May-95.                        
    private int pagelMe1(AdjustableTree tree, int nd) {
        int maxBif = 0;   
        if (!tree.nodeIsInternal(nd)) {
            depth[nd] =0;               //set depth to 0}
            return 0;
        }
        else {
            // node is either unbranched or branched - don't add depth for unbranched internal nodes
            int daughterCount = tree.numberOfDaughtersOfNode(nd);
            if (daughterCount == 1){
                depth[nd] = pagelMe1(tree,tree.firstDaughterOfNode(nd));
                return depth[nd];
            }
            else{  // it branches, so add one to the depth
                for (int daughter=tree.firstDaughterOfNode(nd); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter) ) {
                    int d = pagelMe1(tree,daughter);
                    if (d > maxBif)
                        maxBif = d;
                }
                depth[nd] = maxBif+1;   // this should be compatible with DOS AC3 - a polytomy does add exactly one to the depth (nothing for implied internal ZLB's)
                return depth[nd];
            }
        }
    }
    
       
    /**
     * 
     * This method finishes the transformation (taking
     * polytomies into account.  Note that it does its work while going
     * back down the tree (toward the root), while pagelme1 does its work
     * going up the tree (toward the tips).  Original  9-May-95, Java translation 23-Feb-2001 PEM
     *
     *  @param tree
     *  @param nd
    */
    private void pagelMe2(AdjustableTree tree, int nd) {
        if (tree.nodeIsInternal(nd)) {
            for (int daughter=tree.firstDaughterOfNode(nd); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter) ) { 
                tree.setBranchLength(daughter,depth[nd]- depth[daughter],false);
                pagelMe2(tree,daughter);
            }
        }
    }

	/*.................................................................................................................*/
     public String getName() {
		return "Branch Lengths Method of Pagel (1992) for AC3";
   	 }
	/*.................................................................................................................*/
  	 public String getVersion() {
		return "1.08";
   	 }
	/*.................................................................................................................*/
  	 public boolean isPrerelease() {
		return false;
   	 }
   	 
	/*.................................................................................................................*/
  	 public String getAuthors() {
		return "Franois Gao , Jonathan Corcy, Michel Laurin (taken from PDAP)";
   	 }
	/*.................................................................................................................*/
 	/** returns an explanation of what the module does.*/
 	public String getExplanation() {
 		return ("Adjusts a tree's branch lengths according the method of Pagel (1992): " +
 		             "The depth of a node is equal to the maximum of the number of " +
				"bifurcation levels of the left child and the right child plus one. ");
   	 }

}