/* AC3 Mesquite module */


package mesquite.ac3.lib;
/*~~  */

import java.awt.*;
import java.util.Vector;

import mesquite.lib.*;
//import JSci.maths.statistics.*;    // for p-value calculations

// AC3StatPaks are based on a record structure used in the original Pascal code for DOS AC3. I have
// based three subclasses off of this: one for regular regression (used in old screens 1-8,10,11)
// one for regression through the origin (screen 9) and one for the regression based on root
// reconstruction (DOS screens 9A,B,C).  The last one is clearly the odd one of the three descendent
// classes.
import mesquite.lib.characters.CharacterDistribution;

public abstract class AC3StatPak{

	public static final int REGRESSIONCIPI = 1;
	
	protected int n1 = 0;            // axis selection - this may be an obsolete holdover from pascal code PEM
	protected int n2 = 0;            // axis selection
	protected double min[];
	protected double max[];
	protected int number = 0;        // tips or contrasts
	protected int df;                // nominal degrees of freedom
	protected double mean[];      // means 
	protected double variance[];
    protected double cov = 0.0;
    protected double corr = 0.0;
    protected double rma = 0.0;        //reduced major axis
    protected double rmay = 0.0;       //reduced major axis y intercept
    protected double ma = 0.0;         //major axis
    protected double may = 0.0;        //major axis y intercept
    protected double ls = 0.0;         //least squares slope
    protected double lsy = 0.0;        //least square y intercept
	protected NRtDistribution tStats; //an object for calculating t statistics from jsci[Numerical Recipes?]
	public boolean lsregok;           //true if ok to do regression
	// Add support for CI/PI here, because it isn't inconceivable
	protected Vector CIPIValues = null;
	// Although these are currently only used with RootStats (a AC32CTStatPak), there is nothing
	// necessarily preventing confidence interval stats, so these are moved up here.  PEM 21 June 2006
	protected double 			width1 = 0.95;	        //Confidence/prediction interval width
	protected double              width2 = 0.90;          //Second confidence/prediction interval width
//modif here
	protected Tree theTree;
	protected CharacterDistribution observedStates;
	//end modif
	
    public AC3StatPak() {
 	    min = new double[]  {MesquiteDouble.infinite , MesquiteDouble.infinite};
	    max = new double[] { MesquiteDouble.negInfinite, MesquiteDouble.negInfinite};
	    mean = new double[] {0.0,0.0};
	    variance = new double[] {0.0,0.0};
    }


    /** A common method for reseting the chart bounds, which may get subsequently set in doCalculations */
    protected void resetChartBounds(MesquiteChart chart) {
	    //If it's time to recalculate, then turn off stale chart contraints
	    if (chart!=null && (chart.isMinimumYConstrained() || chart.isMaximumYConstrained())){
	      	chart.deConstrainMinimumX();
		    chart.deConstrainMaximumX();
		    chart.deConstrainMinimumY();
		    chart.deConstrainMaximumY();
		    //chart.deConstrainMinimumZ();
		    //chart.deConstrainMaximumZ();
	    }
    }

    //The following were moved here along with the corresponding fields.
    public void setTree(Tree tree){
    	this.theTree=tree;
    }
    
    public void setObservedStates(CharacterDistribution obs1){
    	this.observedStates=obs1;
    }
    
	public double getWidth1(){
		return width1;
	}
  	
	public double getWidth2(){
		return width2;
	}
  	public void querySetWidth(){
  	}
  	public void setWidth1(double w1){
  		width1 = w1;
  	}
  	public void setWidth2(double w2){
  		width2 = w2;
  	}

  	public double getVh(){
  		return 0;  //??
  	}

  	
  	public void addCIPIValue(MesquiteDouble newValue){
  		if (CIPIValues == null)
  			CIPIValues = new Vector();
  		CIPIValues.add(newValue);
  	}
  	
	public void clearCIPIlist(){
		if (CIPIValues != null)
			CIPIValues.clear();
	}


    /**
    Do basic calculations on number arrays passed.  After these are done, it should be possible to query
    the statpak for results.  Should return whether or not chart needs to be re-munched*/
    public abstract boolean doCalculations(NumberArray xArray, NumberArray yArray, MesquiteChart chart);

 
    // A few simple methods that everyone shares and that provide some encapsulation

    /** returns the number of data points (for df or variance calculations) 
	 @return integer */
    public int getNumber() {
	    return number;
	}

	/** 
	@return one less than the number of data points (for df calculations)  */
	public int getNumberMinus1() {
		return number-1;
	}


	/** what the statpak thinks the degrees of freedom should be (compare getNumber)
	@return degrees of freedom */
	public int getDegreesOfFreedom() {
		return df;
	}

	/** 
	 @return the correlation coefficient*/
	/* Wayne: I think we'll have to settle for protected here, private makes them inaccessable to subclasses.*/
	public double getCorrelationCoefficient(){
		return corr;
	}

	/** 
	 @return the calculated covariance */
	public double getCovariance() {
		return cov;
	}

	/** returns the calculated reduced major axis slope */
	public double getReducedMajorAxis() {
    		return rma;  
	}
 
	/** returns the calculated reduced major axis y intercept */
	public double getReducedMajorAxisYIntercept() {
    		return rmay;  
	}

	/** returns the calculated major axis slope */
	public double getMajorAxis() {
		return rma;  
	}
 
	/** returns the calculated major axis y intercept */
	public double getMajorAxisYIntercept() {
    	return rmay;  
	}
 
	/** returns the standard least squares regression slope */
	public double getLeastSquaresSlope() {
		return ls;
	}

	/** returns the standard least squares y intercept */
	public double getLeastSquaresYIntercept() {
		return lsy;
	}

	/** returns true if regression results are valid (>=2 points, slope not infinite, etc.) */
	public boolean getRegressionOK() {
		return lsregok;
	}

	/** Returns calculated p-value */
	public abstract double getPValue();

	/** Returns calculated variance */
	public double getVariance(int which){
		if (which >= 0 && which < 2)
			return variance[which];
		return MesquiteDouble.unassigned;
	}

	/** Returns calculated mean */
	public double getMean(int which){
		if (which >= 0 && which < 2)
			return variance[which];
		return MesquiteDouble.unassigned;
	}


	// Both the regular regression StatPak takes values directly
	// from the array, rather than off the tree (like the AC32CTStatPak).  So following are 
	// methods that these two share.  The AC32CTStatPak overrides this.

	// sets the min and max fields to the appropriate minima and maxima for the two NumberArrays
	// since min and max are zero based, maybe the names data0 and data1 will be less confusing
	protected void calcBounds(NumberArray data0, NumberArray data1) {
		number = 0;
		for (int i = 0; i < data0.getSize(); i++)
			if (data0.isCombinable(i) && data1.isCombinable(i)) {
				number++;
				if (data0.getDouble(i) > max[0])
					max[0] = data0.getDouble(i);
				if (data1.getDouble(i) > max[1])
					max[1] = data1.getDouble(i);
				if (data0.getDouble(i) < min[0])
					min[0] = data0.getDouble(i);
				if (data1.getDouble(i) < min[1])
					min[1] = data1.getDouble(i);
			}
	}

	// Accessers for the values set in getBounds 
	/** returns the minimum value of the character/trait specfied by index in the data set
	@param which integer- specifies which character/trait
	@return double - the minimum value */
	public double getMin(int which) {
		if (which >= 0 && which < 2)
			return min[which];
		return MesquiteDouble.unassigned;
	}

	/** returns the maximum value of the character/trait specified by index in the data set */
	public double getMax(int which) {
		if (which >= 0 && which < 2)
			return max[which];
		return MesquiteDouble.unassigned;
	}

	// calculates raw sums, sums of squares and cross-product terms.
	protected void getSums(NumberArray data0, 
 						   NumberArray data1,	
 						   MesquiteDouble sx, 
 						   MesquiteDouble sx2,	
 						   MesquiteDouble sy,
						   MesquiteDouble sxy) {
		for (int i = 0; i < data0.getSize(); i++)
			if (data0.isCombinable(i) && data1.isCombinable(i)) {
				sx.add(data0.getDouble(i));
				sx2.add(data0.getDouble(i)*data0.getDouble(i));
				sy.add(data1.getDouble(i));
				sxy.add(data0.getDouble(i)*data1.getDouble(i)); 
		}
	}


	// And one for setting count based on the number of internal nodes in a tree
	public void countInternalNodes(Tree tree, int node) {
		number = tree.numberOfInternalsInClade(node);
	}
	

    
	
	public void clearCIPIList(){
		if (CIPIValues != null)
			CIPIValues.clear();
	}
	
	public abstract void setDrawingOption(int option);

	/** Draws regression lines and other items on the chart. **/
	public abstract void drawOnChart(Graphics g, MesquiteChart chart);
	
	/** Returns results as String **/
	public abstract String flst();

	/** Returns results as brief String for use in legends **/
	public abstract String getLegendText();


}