/* AC3 Mesquite module */

package mesquite.ac3.lib;

import mesquite.lib.*;


public class NoPolyTree extends MesquiteTree {
	MesquiteTree original;

	public NoPolyTree (MesquiteTree tree) {
		super(tree.getTaxa());
		original = tree;
		setToClone(tree);
	}
	public void disconnect(){
		original = null;
	}
	public void notifyListeners(Object caller, Notification notification) { 
		int code = Notification.getCode(notification);
		if (code == MesquiteListener.SELECTION_CHANGED) {
			original.notifyListeners(caller, notification);
		}
		super.notifyListeners(caller, notification);
	}
	/*-----------------------------------------*/
	/** Sets whether or not the part is selected */
	public void setSelected(int part, boolean select) {
		original.setSelected(part, select);
		super.setSelected(part,select);
	}
	/*-----------------------------------------*/
	/** Deselects all parts */
	public void deselectAll(){
		original.deselectAll();
		super.deselectAll();
	}
	/*-----------------------------------------*/
	/** Selects all parts */
	public void selectAll(){
		original.selectAll();
		super.selectAll();
	}
	/*-----------------------------------------*/
	public void setToClone(MesquiteTree tree){
		super.setToClone(tree);
		polysToZeroLength();
	}
	
	/*-----------------------------------------*/
	/** Sets the tree to be a default bush.*/
	public void setToDefaultBush(int numTaxa, boolean notify) {
		super.setToDefaultBush(numTaxa, notify);
		polysToZeroLength();
	}
	
	/*-----------------------------------------*/
	/** Collapses branch to yield polytomy.*/
	public synchronized boolean collapseBranch(int node, boolean notify) {   
		boolean c = super.collapseBranch(node, notify);
		polysToZeroLength();
		return c;
	}
	/*-----------------------------------------*/
	/** Collapses all internal branches within clade above node, to yield bush.*/
	public synchronized boolean collapseAllBranches(int node, boolean below, boolean notify) {   
		boolean c = super.collapseAllBranches(node, below, notify);
		polysToZeroLength();
		return c;
	}
	/*-----------------------------------------*/
	/** Returns a polytomous node.*/
	public  int findPolytomy(int node) {
		if (nodeIsTerminal(node))
			return -1;
		if (nodeIsPolytomous(node))
			return node;
		for (int d = firstDaughterOfNode(node); nodeExists(d); d = nextSisterOfNode(d))
				if (hasPolytomies(d))
					return findPolytomy(d);
		return -1;
	}

	/*-----------------------------------------*/
	private void polysToZeroLength(){
		if (!checkAncestorFunctionEquivalence(original.getRoot()))
			MesquiteMessage.warnProgrammer("error 1: AC3 NoPolyTree doesn't match original");
		while (hasPolytomies(getRoot())){
			int p = findPolytomy(getRoot());
			int d = firstDaughterOfNode(p);
			double origDL = getBranchLength(d);
			int ns = nextSisterOfNode(d);
			double origNSL = getBranchLength(ns);
			moveBranch(d, ns, false);
			setBranchLength(d, origDL, false);
			setBranchLength(ns, origNSL, false);
			setBranchLength(firstDaughterOfNode(p), 0, false);
		}
		if (!checkAncestorFunctionEquivalence(original.getRoot()))
			MesquiteTrunk.mesquiteTrunk.alert("error 2: AC3 NoPolyTree doesn't match original.  Original  " + original.writeTreeWithNodeNumbers() + " noPolyTree " + this.writeTreeWithNodeNumbers());
	}


	/*-----------------------------------------*/
	/** Checks whether this & original tree have save ancestor functions.  To ensure tree node numbering remains equivalent for selection synchronization.*/
	private boolean checkAncestorFunctionEquivalence(int oNode) {
		if (original.nodeIsTerminal(oNode))
			return true;
		for (int d = original.firstDaughterOfNode(oNode); original.nodeExists(d); d = original.nextSisterOfNode(d))
				if (!checkAncestorFunctionEquivalence(d))
					return false;
		if (original.nodeIsPolytomous(oNode))
			return true;
		if (getRoot() == oNode)
			return oNode == original.getRoot();
		int noPolyAncNode = oNode;
		for (int oAncNode = oNode; original.nodeExists(oAncNode); oAncNode = original.motherOfNode(oAncNode), noPolyAncNode = motherOfNode(noPolyAncNode)){
			if (oAncNode!=noPolyAncNode) {
				if (original.nodeIsPolytomous(oAncNode)){ //skip down extra
					while (nodeExists(noPolyAncNode) && noPolyAncNode != oAncNode)
						noPolyAncNode = motherOfNode(noPolyAncNode);
					if (!nodeExists(noPolyAncNode)) 
						return false;
				}
				else 
					return false;
			}
		}
		return true;
	}
	
	public void mapDeletedFromSource(Tree tree,boolean[] newDeleted, boolean[] oldDeleted){
		
	}
	
}