/* Mesquite ~~ Copyright 1997-2006 W. & D. Maddison*/
/* This java translation and commentary copyright 2001-2006 Peter E. Midford */
/* This file copyright 2003-2006 Peter E. Midford */

package mesquite.ac3.lib;

import mesquite.lib.*;
import mesquite.cont.lib.*;


public class SimpleContrastCalculator {
	protected Tree tree;
	protected ContinuousDistribution states;
	protected int numNodes;
	protected double [] contrasts = null;
	protected double [] rawContrasts = null;
	protected double [] x = null;
	protected double [] vPrime = null;
	protected double [] stdDev = null;
	protected double [] height = null;
	protected double [] weightPassThrough = null;   // needed for parents of nodes with missing data
	protected boolean weightedBranches = true;
	private boolean statsOk = true;
	

	/** Constructor  for a simple calculator */
	public SimpleContrastCalculator(Tree theTree, ContinuousDistribution observedStates) {
		super();
		if (theTree == null || observedStates == null) {
			MesquiteMessage.printStackTrace("A null value was passed to the constructor for SimpleContrastCalculator");
			tree = null;
			observedStates = null;
		}
		else {
			tree = theTree;
			states = observedStates;
			numNodes = tree.getNumNodeSpaces();
			checkArrays();
		}	
	}
	
	public boolean Ok () {
		return statsOk;
	}
	
	/** this tries to reuse as much memory as possible, while resetting everything */
	public void reinitialize(Tree theTree, ContinuousDistribution observedStates) {
		tree = theTree; //reinitialize internals
		states = observedStates;
		numNodes = tree.getNumNodeSpaces(); 
		checkArrays(); // either blank all arrays or allocate new ones if space requirements change
	}
	
	/*..................................................................................................................*/
	/** fills or updates the internal arrays */
	protected void checkArrays() {
		if (contrasts == null || contrasts.length != numNodes) {
			contrasts = new double[numNodes];
			rawContrasts = new double[numNodes];
			vPrime = new double[numNodes];
			x = new double[numNodes];
			stdDev = new double[numNodes];
			height = new double[numNodes];
			weightPassThrough = new double[numNodes];
		}
		else {
			DoubleArray.deassignArray(contrasts);
			DoubleArray.deassignArray(rawContrasts);  //should this be zeroArray?
			DoubleArray.zeroArray(x);
			DoubleArray.zeroArray(vPrime);
			DoubleArray.zeroArray(stdDev);
			DoubleArray.deassignArray(weightPassThrough);
		}
	}
	
	/*.................................................................................................................*/
	/** implements the notion of weighted branches for contrast calculations. */
	/** this needs to be reworked for missing data or removed */
	protected double branchWeight (int node){
		double weight;
		if (weightedBranches){
			if (node == tree.getRoot()){  /*asking for weight across root for unrooted case*/
//				int left = tree.firstDaughterOfNode(node);
//				int right = tree.lastDaughterOfNode(node);
//				if (tree.branchLengthUnassigned(left))
//					weight=tree.getBranchLength(right);
//				else if (tree.branchLengthUnassigned(right))
//					weight=tree.getBranchLength(left);
//				else
//					weight=tree.getBranchLength(left) +tree.getBranchLength(right);
				weight = 0.0;
			}
			else if (tree.branchLengthUnassigned(node))   //give warning?
				weight=1.0;
			else
				weight=tree.getBranchLength(node);
			}
		else
			weight=1.0;
			
		//if (weight==0)  //doing this breaks zero-length branch polytomies
		//	weight=1.0;  //give warning?
		return(weight);
	}
	/*.................................................................................................................*/
	// Added checks for missing data and handle the cases of interior nodes with one or both daughters
	// missing data.
	private void vCalculation(int node) {
		for (int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter))
			vCalculation(daughter);
		if (tree.nodeIsInternal(node)) {
			int left = tree.firstDaughterOfNode(node);
			int right = tree.lastDaughterOfNode(node);
			if (MesquiteDouble.isCombinable(vPrime[left]) && MesquiteDouble.isCombinable(vPrime[right]))  {// both good
				vPrime[node] = branchWeight(node) + (vPrime[left] * vPrime[right]) / (vPrime[left] + vPrime[right]); /*vPrime used temporarily!*/
				stdDev[node] = Math.sqrt(vPrime[right] + vPrime[left]);				
			}
			else if (MesquiteDouble.isCombinable(vPrime[left])) { // only left good
					weightPassThrough[node] = branchWeight(left);
					//vPrime[left] += branchWeight(node);
					vPrime[node] = branchWeight(node)+vPrime[left];
					stdDev[node] = MesquiteDouble.unassigned;
				}
			else if (MesquiteDouble.isCombinable(vPrime[right])){ // only right good
					weightPassThrough[node] = branchWeight(right);
					//vPrime[right] += branchWeight(node);
					vPrime[node] = branchWeight(node) + vPrime[right];
					stdDev[node] = MesquiteDouble.unassigned;
				}
			else { // both bad  - so the whole subtree below this node is dead?
					vPrime[node] = MesquiteDouble.unassigned;
					stdDev[node] = MesquiteDouble.unassigned;
			}
		}		
		else { // Check that the node actually has data available
			if (states != null  && 
			     MesquiteDouble.isCombinable(x[node]=states.getState(tree.taxonNumberOfNode(node), 0))) {  
				//x[node] = states.getState(tree.taxonNumberOfNode(node), 0);
				vPrime[node] = branchWeight(node);
			}
			else{  //bad data
				x[node] = MesquiteDouble.unassigned;
				vPrime[node] = MesquiteDouble.unassigned;
			}
		}
	}
	/*......................................................................................................................*/
	protected void heightCalculation(double ht, int node, int mode) {
		double myHeight;
		if (tree.nodeIsInternal(node)) {
			if (MesquiteDouble.isCombinable(stdDev[node])) { // good data
				if (node != tree.getRoot())
					if (mode == ContrastForChar.RAWHEIGHT)
						myHeight = ht + branchWeight(node);
					else
						myHeight = ht + vPrime[node];
				else    
					myHeight = 0;
				height[node] = myHeight;
			}
			else {
				height[node] = MesquiteDouble.unassigned;
				if (mode == ContrastForChar.RAWHEIGHT)
					myHeight = ht + branchWeight(node);
				else
					myHeight = ht + branchWeight(node);  //vPrime[node];
			}
			for (int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter))
				heightCalculation(myHeight, daughter,mode);
		}
		else {  // a tip
			if (MesquiteDouble.isCombinable(x[node])) {
				if (mode == ContrastForChar.RAWHEIGHT)
					height[node] = ht + branchWeight(node);
				else
					height[node] = ht + vPrime[node];
			}
			else 
				height[node] = MesquiteDouble.unassigned;		
		}
	}

	/*.................................................................................................................*/
	private void xCalculation(int node) {
		for (int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter))
			xCalculation(daughter);
		if (tree.nodeIsInternal(node)) {
			int left = tree.firstDaughterOfNode(node);
			int right = tree.lastDaughterOfNode(node);
			if (MesquiteDouble.isCombinable(x[left]) && MesquiteDouble.isCombinable(x[right])) {  // good contrast
				x[node] = (vPrime[right]*x[left] + vPrime[left]*x[right] )/(vPrime[left] +vPrime[right]);  
				rawContrasts[node] =(x[left] - x[right]);
				contrasts[node] = rawContrasts[node]/stdDev[node];
			}
			else {
				rawContrasts[node] = MesquiteDouble.unassigned;
				contrasts[node] = MesquiteDouble.unassigned;
				x[node] = MesquiteDouble.unassigned;  // Default unless either one of left or right is good
				if(MesquiteDouble.isCombinable(x[right]))
					x[node] = x[right];  // no value from left to average with, so all we can do is propagate down!
				if(MesquiteDouble.isCombinable(x[left]))
					x[node] = x[left];  // no value from right to average with, so all we can do is propagate down!
			}
		}
		else {
			if (states != null  && MesquiteDouble.isCombinable(states.getState(tree.taxonNumberOfNode(node), 0)))  
				x[node] = states.getState(tree.taxonNumberOfNode(node), 0); 
			else
				x[node] = MesquiteDouble.unassigned;
			stdDev[node] = MesquiteDouble.unassigned;
			contrasts[node] = MesquiteDouble.unassigned;
		}
	}
	
	/** This returns the array of contrast values for those modules that need them 
	@return double[] - the array of contrast values calculated here */
	public double [] contrastCalculation() {
		int root = tree.getRoot();
		vCalculation(root);
		xCalculation(root);
		return contrasts;
	}	

	/** This returns the array of contrast values for those modules that need them 
	@return double[] - the array of raw contrast values calculated here or null for failure */
	public double [] rawContrastCalculation() {
		int root = tree.getRoot();
		vCalculation(root);
		xCalculation(root);
		return rawContrasts;
	}
	

	/** This returns the array of contrast values for those modules that need them 
	@return double[] - the array of raw contrast values calculated here or null for failure */
	public double [] nodeValueCalculation() {
		int root = tree.getRoot();
		vCalculation(root);
		xCalculation(root);
		return x;
	}
	

	/** This returns the array of contrast values for those modules that need them 
	@return double[] - the array of raw contrast values calculated here or null for failure */
	public double [] vPrimeCalculation() {
		int root = tree.getRoot();
		vCalculation(root);
		return vPrime;
	}


	/** This returns the array of contrast values for those modules that need them 
	@return double[] - the array of raw contrast values calculated here or null for failure */
	public double [] stdDevCalculation() {
		int root = tree.getRoot();
		vCalculation(root);
		return stdDev;
	}

	/** This copies tip values into the x array 
	@return double [] - the array of tip values, internal node values are set to MesquiteDouble.unassigned */
	public double [] fillTips(int node) {
		for (int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter))
			fillTips(daughter);
		if (tree.nodeIsInternal(node)) {
			x[node] = MesquiteDouble.unassigned;
		}
		else {
			x[node] = states.getState(tree.taxonNumberOfNode(node), 0);  
		}
		return x;
	}

	/** This also assumes things have been set up. */
	public double [] heightCalculation(int mode){
		int root = tree.getRoot();
		vCalculation(root);
		heightCalculation(0.0,root,mode);
		return height;
	}
	
	/** Just returns the standard deviation array, without recalculating */
	public double [] getStdDev() {
		return stdDev;
	}

	/** Just returns the vPrime array, without recalculating */
	public double [] getVPrime() {
		return vPrime;
	}
	
	
	/** Just returns the node values (tip and interior), without recalculating */
	public double [] getNodeValues() {
		return x;
	}
	
	
}	